# imports for python
import numpy as np
import math
import warnings
from typing import List, Union

# Replacement for libc.math and cython types
from math import fabs, pow, exp, log, log1p

INFINITY = float("inf")

ArrayLike = Union[np.ndarray, List[float]]

"""
Notation:

    FPR = false positive rate
    FNR = false negative rate
    alpha = another name for FPR
    beta = another name for FNR
    order = Renyi order
    epsilon = value of Renyi divergence


The main python facing function is get_FNR.
This function takes as input an single
(order, epsilon)-RDP guarantee and a single (or multiple)
FPR values alpha and returns the corresponding FNR values of
the tradeoff curve

The function does this by calling one of 3 cython functions,
each optimized for a particular regime of Renyi order (order).
They are:

get_FNR_case1_vec: handles order > 1
get_FNR_case2_vec: handles order = 1
get_FNR_case3_vec: handles 0 < order < 1

Each of these functions perform a for loop (in cython) over
the list/array of given FPR values. For each FPR value, the functions
call a correspinding _core_logic function, which implements the actual
conversion. The conversion is itself a root-finding problem, and various
numerical techniques are used to speed up the compuation. The main doc
string that explains these techniques are in _core_logic_case1.
The doc strings of all other core logic functions only mention techniques
different from those in _core_logic_case1.
"""


###############
###############
### CASE 1 ####
###############
###############
def get_fixed_point_case1(order, upper, tol=1e-7):
    """
    See _core_logic_case1 doc string for context on why we need this
    function.

    Assuming order > 1, this function returns the fixed point of the function beta(alpha)
    implicitly defined via:

    (1-alpha)^{1-order} beta^order     + alpha^{1-order} (1-beta)^order     <= upper
    (1-alpha)^{order}   beta^{1-order} + alpha^{order}   (1-beta)^{1-order} <= upper

    which can be easily seen to be the smallest value of beta such that:

    (1-beta)^{order} beta^{1-order} + beta^{order} (1-beta)^{1-order} <= upper

    Let the lhs of this equation by F(beta). F can be shown to be convex and
    decreasing. Since we know this function computes the fixed point of a
    tradeoff function, we know that 0 <= beta <= 1/2.

    Therefore, this function uses bisection over [0,1/2] to find F(beta) >= upper.
    Let beta* denote the true inverse of F at upper. It is guaranteed that
    beta* < beta < beta* + tol thanks to bisection. Since this function is not called
    as often as _core_logic_case1, it is not as performance critical, and we
    intentionally use standard bisection.
    """

    # init [low, high]  for bisection
    low = 0
    high = 1 / 2

    # init all other vars
    mid = 0.0
    F_mid = 0.0

    # run bisection
    while high - low > tol:
        mid = 0.5 * (low + high)
        F_mid = (1 - mid) ** (1 - order) * mid**order + mid ** (1 - order) * (
            1 - mid
        ) ** order

        if F_mid > upper:  # True means "move low up"
            low = mid
        else:  # False means "move high down"
            high = mid

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def _core_logic_case1(alpha, upper, order, one_minus_order, fixed_point, tol):
    """
    For fixed RDP guarantee (order, epsilon) with order > 1 and FPR alpha,
    this function returns the smallest value of beta in [0,1] that simultaneously satisfies:

    (1-alpha)^{1-order} beta^order     + alpha^{1-order} (1-beta)^order     <= upper
    (1-alpha)^{order}   beta^{1-order} + alpha^{order}   (1-beta)^{1-order} <= upper

    Let lhs1 denote the lhs of the first inequality, and lhs2 denote the second.
    Let M(beta) = max(lhs1, lhs2). The above two inequalities simplify to M(beta) <= upper.
    Let fixed_point denote the unique value such that (alpha = fixed_point, beta = fixed_point)
    satisfies the two inequalities above. It can be shown that M(beta) = lhs1(beta) if alpha < fixed_point
    and M(beta) = lhs2(beta) if alpha > fixed_point. The code below uses this fact to focus on
    only one of the two inequalities above, depending on the value of FPR alpha.

    Note that if alpha = 0, then only beta = 1 satisfies the inequalities. And if alpha = 1, then all values
    of beta satisfy the inequalities, therefore we choose the smallest: beta = 0. These two edge
    cases are hard coded into the function.

    M(beta) is a convex decreasing function of beta, and our goal is to return the smallest
    value of beta such that M(beta) <= upper. Let the result of this be denoted beta = f(alpha).
    Since we know this is a tradeoff curve, we have that 0 <= beta <= 1 - alpha.
    Therefore, we can use bisection over [0, 1-alpha] to numerically invert M(beta) <= upper.

    Let beta* denote the true inverse of M at upper, i.e. M(beta*) = upper. Let beta denote the
    output of this function. This function is guaranteed to return beta such that
    beta* < beta < beta* + tol.

    This function implements one more trick for speed: since M is convex and decreasing,
    using Newton's Method is guaranteed to yield a lower bound on beta*, and using the secant
    method is guaranteed to yield an upper bound on beta*. Therefore, we use both techniques to
    improve the bisection interval when possible.

    Args:
        alpha: FPR value
        upper: exp( (order-1) * epsilon). The value of the upper bound
        order: Renyi order
        one_minus_order: one minus the Renyi order
        fixed_point: fixed point of the tradeoff curve
        tol: tolerance. See doc string for meaning

    Returns:
        beta: FNR at FPR
    """

    # handle alpha = 0 or alpha = 1 explicitly
    if alpha == 0:
        return 1.0
    elif alpha == 1:
        return 0.0

    # save variables to reduce computation of M for later
    one_minus_alpha = 1.0 - alpha
    c1 = 0.0
    c2 = 0.0
    expo = 0.0

    if alpha <= fixed_point:
        expo = order
        c1 = pow(one_minus_alpha, one_minus_order)
        c2 = pow(alpha, one_minus_order)
    else:
        expo = one_minus_order
        c1 = pow(one_minus_alpha, order)
        c2 = pow(alpha, order)

    # init [low, high] and M(low), M(high) for bisection
    low = 0.0
    high = one_minus_alpha

    # M_low, M_high below assume low = 0 and high = 1 - alpha
    # if alpha <= fixed_point, expo > 0 and M_low is finite, if expo < 0, M_low is infinite
    M_low = c2 if alpha <= fixed_point else INFINITY

    # M_high: cython.double = c1 * pow(high, expo) + c2 * pow(1 - high, expo)
    M_high = c1 * pow(high, expo) + c2 * pow(alpha, expo)

    # init all other vars for bisection
    mid = 0.5 * (high - low)
    one_minus_mid = 0.0
    term1 = 0.0
    term2 = 0.0
    M_mid = 0.0
    grad = 0.0
    newton = 0.0
    secant = 0.0

    # begin bisection
    while (high - low) > tol:
        one_minus_mid = 1.0 - mid

        # M(mid) = term1 + term2
        term1 = c1 * pow(mid, expo)  # used in Newton calc
        term2 = c2 * pow(one_minus_mid, expo)  # used in Newton calc
        M_mid = term1 + term2

        # update bisection interval [low, high]
        if M_mid > upper:  # True means "move low up"
            low = mid
            M_low = M_mid  # used in Secant calc
        else:  #  False means "move high down"
            high = mid
            M_high = M_mid  # used in Secant calc

        # attempt to use Newton's method to increase low
        grad = expo * (term1 / mid - term2 / one_minus_mid)
        newton = mid - (M_mid - upper) / grad

        # if newton point fits inside bisection interval [low, high],
        # then we know that low < newton < beta* < high, therefore we
        # let low = newton to improve bisection interval.
        if low < newton < high:
            low = newton
            M_low = c1 * pow(low, expo) + c2 * pow(1.0 - low, expo)

        # attempt to use Secant method to decrease high
        if fabs(M_high - M_low) < 1e-15:  # avoid division by 0
            secant = low + (upper - M_low) * (high - low) / (M_high - M_low)

            # if secant point fits inside bisection interval [low, high],
            # then we know that low < beta* < secant < high, therefore we
            # let high = secant to improve bisection interval.
            if low < secant < high:
                high = secant
                M_high = c1 * pow(high, expo) + c2 * pow(1.0 - high, expo)

        # update mid for next iteration of bisection
        mid = 0.5 * (low + high)

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def get_FNR_case1_vec(alpha_array, order, epsilon, tol=1e-7):

    # compute upper and one_minus_order once to save compute in core_logic
    log_upper = (order - 1.0) * epsilon
    upper = exp(log_upper)
    one_minus_order = 1.0 - order

    # compute fixed point (hard code 1e-15 for stability)
    fixed_point = get_fixed_point_case1(order, upper, tol=1e-15)

    # allocate output array
    n = alpha_array.shape[0]
    result_np = np.empty(n, dtype=np.float64)

    # run core logic in C for loop
    for i in range(n):
        result_np[i] = _core_logic_case1(
            alpha_array[i], upper, order, one_minus_order, fixed_point, tol
        )

    return result_np


###############
###############
### CASE 2 ####
###############
###############
def get_fixed_point_case2(epsilon, tol=1e-7):
    """
    See _core_logic_case1 and get_fixed_point_case1 doc string for context on
    why we need this function. Only difference from get_fixed_point_case1 is
    that here, the fixed point equation is:

    beta log( beta/(1-beta) ) + (1-beta) log( (1-beta)/beta) <= epsilon

    The same bisection algo used in get_fixed_point_case1 is used here
    """

    # init [low, high]  for bisection
    low = 0
    high = 1 / 2

    # init all other vars
    mid = 0.0
    F_mid = 0.0

    # run bisection
    while high - low > tol:
        mid = 0.5 * (low + high)
        F_mid = mid * (log(mid) - log1p(-mid)) + (1 - mid) * (log1p(-mid) - log(mid))
        if F_mid > epsilon:  # True means "move low up"
            low = mid
        else:  # False means "move high down"
            high = mid

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def _core_logic_case2(alpha, epsilon, fixed_point, tol):
    """
    See _core_logic_case1 doc string for details. The only difference
    between case1 and this function is that here, order = 1,
    where the inequalities are

    beta log( beta/(1-alpha) ) + (1-beta) log( (1-beta)/alpha) <= epsilon
    alpha log( alpha/(1-beta) ) + (1-alpha) log( (1-alpha)/beta) <= epsilon

    Still, M(beta) defined as in _core_logic_case1 is convex and decreasing,
    so the implementation of bisection + Newton + Secant is the same.
    """
    # handle alpha = 0 or alpha = 1 explicitly
    if alpha == 0:
        return 1.0
    elif alpha == 1:
        return 0.0

    # save variables to reduce computation of M for later
    one_minus_alpha = 1.0 - alpha
    log_one_minus_alpha = log1p(-alpha)
    log_alpha = log(alpha)

    less_than_fixed_point = alpha <= fixed_point

    # init [low, high] and M(low), M(high) for bisection
    low = 0.0
    high = one_minus_alpha

    M_low = 0.0
    M_high = 0.0

    if less_than_fixed_point:
        # M_low = M(0) = -log(alpha) (limit as beta -> 0+)
        M_low = -log_alpha
        M_high = high * (log(high) - log_one_minus_alpha) + (1.0 - high) * (
            log1p(-high) - log_alpha
        )
    else:
        # M(0) = +inf (because -(1-alpha)*log(beta) term blows up)
        M_low = INFINITY
        M_high = alpha * (log_alpha - log1p(-high)) + one_minus_alpha * (
            log_one_minus_alpha - log(high)
        )

    # begin bisection
    mid = 0.5 * (low + high)
    one_minus_mid = 0.0
    M_mid = 0.0
    grad = 0.0
    newton = 0.0
    secant = 0.0

    while (high - low) > tol:
        one_minus_mid = 1.0 - mid

        # compute M(mid) and M'(mid)
        if less_than_fixed_point:
            M_mid = mid * (log(mid) - log_one_minus_alpha) + one_minus_mid * (
                log1p(-mid) - log_alpha
            )
            grad = (log(mid) - log_one_minus_alpha) - (log1p(-mid) - log_alpha)
        else:
            M_mid = alpha * (log_alpha - log1p(-mid)) + one_minus_alpha * (
                log_one_minus_alpha - log(mid)
            )
            grad = (mid - one_minus_alpha) / (mid * one_minus_mid)

        # update bisection interval [low, high]
        if M_mid > epsilon:  # True means "move low up"
            low = mid
            M_low = M_mid  # used in Secant calc
        else:  #  False means "move high down"
            high = mid
            M_high = M_mid  # used in Secant calc

        # attempt to use Newton's method to increase low
        newton = mid - (M_mid - epsilon) / grad

        # if newton point fits inside bisection interval [low, high],
        # then we know that low < newton < beta* < high, therefore we
        # let low = newton to improve bisection interval.
        if low < newton < high:
            low = newton

            # recompute M_low
            if less_than_fixed_point:
                M_low = low * (log(low) - log_one_minus_alpha) + (1.0 - low) * (
                    log1p(-low) - log_alpha
                )
            else:
                M_low = alpha * (log_alpha - log1p(-low)) + one_minus_alpha * (
                    log_one_minus_alpha - log(low)
                )

        # attempt to use Secant method to decrease high
        if fabs(M_high - M_low) < 1e-15:  # avoid division by 0
            secant = low + (epsilon - M_low) * (high - low) / (M_high - M_low)

            # if secant point fits inside bisection interval [low, high],
            # then we know that low < beta* < secant < high, therefore we
            # let high = secant to improve bisection interval.
            if low < secant < high:
                high = secant

                # recompute M_high
                if less_than_fixed_point:
                    M_high = high * (log(high) - log_one_minus_alpha) + (1.0 - high) * (
                        log1p(-high) - log_alpha
                    )
                else:
                    M_high = alpha * (log_alpha - log1p(-high)) + one_minus_alpha * (
                        log_one_minus_alpha - log(high)
                    )

        # Next bisection midpoint
        mid = 0.5 * (low + high)

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def get_FNR_case2_vec(alpha_array, epsilon, tol=1e-7):

    # compute fixed point (hard code 1e-15 tol for stability)
    fixed_point = get_fixed_point_case2(epsilon, tol=1e-15)

    # allocate output array
    n = alpha_array.shape[0]
    result_np = np.empty(n, dtype=np.float64)

    # run core_logic in C for loop
    for i in range(n):
        result_np[i] = _core_logic_case2(alpha_array[i], epsilon, fixed_point, tol)

    return result_np


###############
###############
### CASE 3 ####
###############
###############
def get_fixed_point_case3(order, lower, alpha_max, tol=1e-7):
    """
    See _core_logic_case1 and get_fixed_point_case1 doc string for context on
    why we need this function. Only difference from get_fixed_point_case1 is
    that here, the fixed point equation is:

    (1-beta)^{order} beta^{1-order} + beta^{order} (1-beta)^{1-order} >= lower

    Since order < 1 for case 3, now F(beta) (the lhs of the inequality) is
    increasing and concave, as opposed to decreasing and convex as in case 1.
    Moreover, in get_fixed_point_case1 all we knew was the the solution had
    to be between [0,1/2]. Here, we know that the fixed point must be less than
    alpha_max, so we search between [0,alpha_max].

    The code below returns a value of beta that is guaranteed to have the property
    F(beta) >= lower. It does this by bisection. Let beta* denote the true inverse of
    F at lower. It is guaranteed that beta* < beta < beta* + tol.
    """

    # init [low, high]  for bisection
    low = 0
    high = alpha_max

    # init all other vars
    mid = 0.0
    F_mid = 0.0

    # run bisection
    while high - low > tol:
        mid = (low + high) / 2.0
        F_mid = (1 - mid) ** (1 - order) * mid**order + mid ** (1 - order) * (
            1 - mid
        ) ** order
        if F_mid > lower:
            high = mid  # True means "move high down"
        else:
            low = mid  # False means "move low up"

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def _core_logic_case3(
    alpha, lower, order, one_minus_order, fixed_point, alpha_max, tol
):
    """
    See _core_logic_case1 doc string for details. The only difference
    between case 1 and this function is that here, order < 1,
    where the inequalities are

    (1-alpha)^{1-order} beta^order     + alpha^{1-order} (1-beta)^order     >= lower
    (1-alpha)^{order}   beta^{1-order} + alpha^{order}   (1-beta)^{1-order} >= lower

    In particular, they are identical except with the inequality fliped.
    This has a much bigger impact than one may imagine. For example, M(beta) is
    now concave and increasing since we need to take the min(lhs1, lhs2) and
    not the max. Note that in case 1 and 2, M(beta) is convex and decreasing.
    More importantly, however, is that for order < 1, the inequalities
    have very different behaviors at alpha = 0.

    At alpha = 0, the inequality simplifies to beta >= e^{ (order-1) / order * epsilon }.
    Therefore, the smallest beta must be e^{ (order-1) / order * epsilon }, which is not 1.

    Due to symmetry, the same effect occurs at beta = 0, in that we get
    alpha >= e^{ (order-1) / order * epsilon }. This leads to the somewhat counter intuitive
    situation where if the input alpha satisfies alpha >= e^{ (order-1) / order * epsilon }, then
    it must be that beta = 0. This is hard coded into the code below (alpha_max is this value).

    The above implies the tradeoff function for RDP when order < 1 has the property
    f(0) < 1 and f(alpha') = 0 for alpha' in some non zero interval that contains 1. This
    implies this tradeoff function catastrophically fails (i.e. has infinite privacy loss).
    We leave further exploration of this to future work.
    """

    # handle alpha = 0 and alpha >= alpha_max explicitly
    if alpha == 0:
        return alpha_max
    elif alpha >= alpha_max:
        return 0.0

    # save variables to reduce computation of M for later
    one_minus_alpha = 1.0 - alpha
    c1 = 0.0
    c2 = 0.0
    expo = 0.0

    if alpha <= fixed_point:
        expo = order
        c1 = pow(one_minus_alpha, one_minus_order)
        c2 = pow(alpha, one_minus_order)
    else:
        expo = one_minus_order
        c1 = pow(one_minus_alpha, order)
        c2 = pow(alpha, order)

    # init [low, high] and M(low), M(high) for bisection
    low = 0.0
    high = one_minus_alpha

    # two lines below assume low = 0 and high = 1 - alpha
    M_low = 0
    M_high = c1 * pow(high, expo) + c2 * pow(alpha, expo)

    # init all other vars for bisection
    mid = 0.5 * (high - low)
    one_minus_mid = 0.0
    term1 = 0.0
    term2 = 0.0
    M_mid = 0.0
    grad = 0.0
    newton = 0.0
    secant = 0.0

    # begin bisection
    while (high - low) > tol:
        one_minus_mid = 1.0 - mid

        # M(mid) = term1 + term2
        term1 = c1 * pow(mid, expo)  # used in Newton calc
        term2 = c2 * pow(one_minus_mid, expo)  # used in Newton calc
        M_mid = term1 + term2

        # update bisection interval [low, high]
        if M_mid > lower:  # True means "move high down"
            high = mid
            M_high = M_mid  # used in Secant calc
        else:  #  False means "move low up"
            low = mid
            M_low = M_mid  # used in Secant calc

        # attempt to use Newton's method to increase low
        grad = expo * (term1 / mid - term2 / one_minus_mid)
        newton = mid - (M_mid - lower) / grad

        # if newton point fits inside bisection interval [low, high],
        # then we know that low < newton < beta* < high, therefore we
        # let low = newton to improve bisection interval.
        if low < newton < high:
            low = newton
            M_low = c1 * pow(low, expo) + c2 * pow(1.0 - low, expo)

        # attempt to use Secant method to decrease high
        if fabs(M_high - M_low) < 1e-15:  # avoid division by 0
            secant = low + (lower - M_low) * (high - low) / (M_high - M_low)

            # if secant point fits inside bisection interval [low, high],
            # then we know that low < beta* < secant < high, therefore we
            # let high = secant to improve bisection interval.
            if low < secant < high:
                high = secant
                M_high = c1 * pow(high, expo) + c2 * pow(1.0 - high, expo)

        # update mid for next iteration of bisection
        mid = 0.5 * (low + high)

    # return high for formal guarantee that beta* < beta < beta* + tol
    return high


def get_FNR_case3_vec(alpha_array, order, epsilon, tol=1e-7):

    # compute lower, one_minus_order, alpha_max once to save compute in core_logic
    log_lower = (order - 1.0) * epsilon
    lower = exp(log_lower)
    one_minus_order = 1.0 - order
    alpha_max = exp((order - 1) / order * epsilon)

    # compute fixed point (hard code 1e-15 for stability)
    fixed_point = get_fixed_point_case3(order, lower, alpha_max, tol=1e-15)

    # allocate output array
    n = alpha_array.shape[0]
    result_np = np.empty(n, dtype=np.float64)

    # run core logic in C for loop
    for i in range(n):
        result_np[i] = _core_logic_case3(
            alpha_array[i], lower, order, one_minus_order, fixed_point, alpha_max, tol
        )

    return result_np


def get_FNR(
    alpha_array: Union[ArrayLike, float],
    order: float,
    epsilon: float,
    tol: float = 1e-7,
):
    """
    Args:
        alpha_array: input FPR(s). Can be a scalar or array like.
        order: Renyi order
        epsilon: Value that upper bounds the Renyi divergence
        tol: Tolerance for the returned FNR.

    Returns:
        FNRs at corresponding FPRs. Same shape as alpha_array.
    """

    # check if epsilon is valid
    if not (epsilon > 0):
        raise ValueError(f"epsilon must be positive, got {epsilon}")

    # check if FPRs alpha are valid, and convert alpha to contiguous 1D numpy array
    is_scalar = np.isscalar(alpha_array)
    alpha_array = np.ascontiguousarray(np.atleast_1d(alpha_array))

    if not (np.all(0 <= alpha_array) and np.all(alpha_array <= 1)):
        raise ValueError(f"FPRs alpha must be between 0 and 1, got {alpha_array}")

    # exp((order - 1.0) * epsilon) is used in all functions; guard against overflow.
    log_val = (order - 1.0) * epsilon
    if log_val >= 200.0:
        warnings.warn(
            f"(order - 1) * epsilon = {log_val:.1f} >= 200: exp would overflow. "
            f"The RDP bound is vacuously loose; returning 0 for all FPR values.",
            RuntimeWarning,
            stacklevel=2,
        )
        result = np.zeros(len(alpha_array), dtype=np.float64)
        return float(result[0]) if is_scalar else result

    # 0 < order < 0.5 are not needed by symmetry
    # per Zhu et al https://arxiv.org/pdf/2106.08567
    if order < 0.5:
        warnings.warn(
            f"RDP order < 0.5 is not needed by symmetry. " f"Got order = {order}"
        )

    # make sure order is valid for the Renyi divergence
    if not (0 < order):
        raise ValueError(f"RDP order must be positive, got {order}")

    if order > 1:
        output = get_FNR_case1_vec(alpha_array, order, epsilon, tol)
    elif order == 1:
        output = get_FNR_case2_vec(alpha_array, epsilon, tol)
    elif order < 1:
        output = get_FNR_case3_vec(alpha_array, order, epsilon, tol)

    if is_scalar:  # if original input alpha_array was a scalar
        return float(output[0])
    else:  # original alpha_array was ArrayLike
        return output



"""DEPRECATED: Use riskcal.analysis instead.

This module is deprecated and will be removed in v2.0.0.

For PLRV functionality, use:
    from riskcal.analysis import PLRVs, get_beta_from_plrvs
"""

import warnings

warnings.warn(
    "The 'riskcal.plrv' module is deprecated and will be removed in v2.0.0. "
    "Use 'riskcal.analysis' instead.",
    DeprecationWarning,
    stacklevel=2,
)

# Re-export from new locations for backward compatibility
from riskcal.analysis._plrv import PLRVs, get_beta

__all__ = [
    "PLRVs",
    "get_beta",
]



"""DP-SGD specific calibration using direct PLD composition (fast path)."""

from typing import Union
import warnings
import numpy as np
from dp_accounting.pld import privacy_loss_distribution as pld_module

from riskcal.calibration.core import (
    PrivacyEvaluator,
    PrivacyMetrics,
    CalibrationTarget,
    CalibrationConfig,
    CalibrationResult,
    calibrate_parameter,
)


def create_dpsgd_evaluator(
    sample_rate: float,
    num_steps: int,
    grid_step: float = 1e-4,
    target_alpha: float | None = None,
) -> PrivacyEvaluator:
    """
    Create a privacy evaluator for DP-SGD.

    Returns a function that maps noise_multiplier → PrivacyMetrics for DP-SGD
    with the specified parameters.

    Args:
        sample_rate: Poisson sampling rate (typically batch_size / dataset_size).
        num_steps: Number of DP-SGD steps (typically num_epochs * steps_per_epoch).
        grid_step: Discretization interval for PLD computation.
        target_alpha: If provided, evaluator will compute beta at this alpha value.
            Required for err_rates calibration.

    Returns:
        PrivacyEvaluator function that computes metrics for given noise_multiplier.

    Example:
        >>> evaluator = create_dpsgd_evaluator(sample_rate=0.002, num_steps=10000)
        >>> metrics = evaluator(noise_multiplier=1.0)
        >>> print(f"Advantage: {metrics.advantage:.4f}")
    """
    from riskcal.analysis import get_advantage_from_pld, get_beta_from_pld

    def evaluator(noise_multiplier: float) -> PrivacyMetrics:
        """Evaluate DP-SGD privacy metrics for given noise."""
        # Create PLD for single step
        pld = pld_module.from_gaussian_mechanism(
            standard_deviation=noise_multiplier,
            sensitivity=1.0,
            value_discretization_interval=grid_step,
        )

        # Compose for num_steps with Poisson sampling
        pld = pld_module.from_gaussian_mechanism(
            standard_deviation=noise_multiplier,
            sampling_prob=sample_rate,
            use_connect_dots=True,
            value_discretization_interval=grid_step,
        ).self_compose(num_steps)

        # Compute metrics
        advantage = get_advantage_from_pld(pld)

        metrics = PrivacyMetrics(advantage=advantage)

        # Compute beta if target_alpha specified (for err_rates calibration)
        if target_alpha is not None:
            beta = get_beta_from_pld(pld, alpha=target_alpha)
            metrics.alpha = target_alpha
            metrics.beta = beta

        return metrics

    return evaluator


def create_dpsgd_epsilon_evaluator(
    sample_rate: float,
    num_steps: int,
    target_delta: float,
    grid_step: float = 1e-4,
) -> PrivacyEvaluator:
    """
    Create an epsilon-delta evaluator for DP-SGD.

    Returns a function that maps noise_multiplier → PrivacyMetrics with epsilon
    for the specified delta.

    Args:
        sample_rate: Poisson sampling rate.
        num_steps: Number of DP-SGD steps.
        target_delta: Delta parameter for epsilon computation.
        grid_step: Discretization interval for PLD computation.

    Returns:
        PrivacyEvaluator that computes epsilon at target_delta.
    """

    def evaluator(noise_multiplier: float) -> PrivacyMetrics:
        """Evaluate epsilon-delta for given noise."""
        pld = pld_module.from_gaussian_mechanism(
            standard_deviation=noise_multiplier,
            sampling_prob=sample_rate,
            use_connect_dots=True,
            value_discretization_interval=grid_step,
        ).self_compose(num_steps)

        epsilon = pld.get_epsilon_for_delta(target_delta)

        return PrivacyMetrics(epsilon=epsilon, delta=target_delta)

    return evaluator


def find_noise_multiplier_for_advantage(
    advantage: float,
    sample_rate: float,
    num_steps: int,
    grid_step: float = 1e-4,
    advantage_tol: float = 1e-3,
    noise_min: float = 0.1,
    noise_max: float = 50.0,
    # Deprecated parameter names for backward compatibility
    advantage_error: float = None,
    mu_error: float = None,
    mu_min: float = None,
    mu_max: float = None,
) -> float:
    """
    Calibrate DP-SGD noise to target advantage (legacy interface).

    .. deprecated:: 1.2.0
        Legacy parameter names (advantage_error, mu_error, mu_min, mu_max)
        will be removed in version 2.0.0.

    Finds minimum noise_multiplier such that advantage ≤ target.

    Args:
        advantage: Target advantage bound in [0, 1].
        sample_rate: Poisson sampling rate.
        num_steps: Number of DP-SGD steps.
        grid_step: PLD discretization interval.
        advantage_tol: Convergence tolerance for advantage.
        noise_min: Lower bound for search.
        noise_max: Upper bound for search.

    Returns:
        Calibrated noise_multiplier (float).

    Example:
        >>> noise = find_noise_multiplier_for_advantage(
        ...     advantage=0.1,
        ...     sample_rate=0.002,
        ...     num_steps=10000
        ... )
        >>> print(f"Use noise_multiplier: {noise:.3f}")

    Note:
        For new code, consider using the generic interface:
        >>> from riskcal.calibration import calibrate_parameter, CalibrationTarget
        >>> evaluator = create_dpsgd_evaluator(sample_rate=0.002, num_steps=10000)
        >>> target = CalibrationTarget(kind='advantage', advantage=0.1)
        >>> result = calibrate_parameter(evaluator, target)
    """
    # Handle backward compatibility for deprecated parameter names
    if advantage_error is not None:
        warnings.warn(
            "Parameter 'advantage_error' is deprecated and will be removed in v2.0.0. "
            "Use 'advantage_tol' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        advantage_tol = advantage_error
    elif mu_error is not None:
        warnings.warn(
            "Parameter 'mu_error' is deprecated and will be removed in v2.0.0. "
            "Use 'advantage_tol' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        advantage_tol = mu_error
    elif advantage_tol is None:
        advantage_tol = 1e-3

    if mu_min is not None:
        warnings.warn(
            "Parameter 'mu_min' is deprecated and will be removed in v2.0.0. "
            "Use 'noise_min' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        noise_min = mu_min

    if mu_max is not None:
        warnings.warn(
            "Parameter 'mu_max' is deprecated and will be removed in v2.0.0. "
            "Use 'noise_max' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        noise_max = mu_max

    evaluator = create_dpsgd_evaluator(
        sample_rate=sample_rate, num_steps=num_steps, grid_step=grid_step
    )

    target = CalibrationTarget(kind="advantage", advantage=advantage)

    config = CalibrationConfig(
        param_min=noise_min,
        param_max=noise_max,
        target_tol=advantage_tol,
        increasing=False,
    )

    result = calibrate_parameter(
        evaluator, target, config, parameter_name="noise_multiplier"
    )

    return result.parameter_value


def find_noise_multiplier_for_err_rates(
    alpha: float,
    beta: float,
    sample_rate: float,
    num_steps: int,
    grid_step: float = 1e-4,
    beta_tol: float = 1e-3,
    noise_min: float = 0.1,
    noise_max: float = 50.0,
    # Deprecated parameter names for backward compatibility
    beta_error: float = None,
    mu_error: float = None,
    mu_min: float = None,
    mu_max: float = None,
) -> float:
    """
    Calibrate DP-SGD noise to target error rates (legacy interface).

    .. deprecated:: 1.2.0
        Legacy parameter names (beta_error, mu_error, mu_min, mu_max)
        will be removed in version 2.0.0.

    Finds minimum noise_multiplier such that beta(alpha) ≤ target_beta.

    Args:
        alpha: Target false positive rate (FPR) in [0, 1].
        beta: Target false negative rate (FNR) in [0, 1].
        sample_rate: Poisson sampling rate.
        num_steps: Number of DP-SGD steps.
        grid_step: PLD discretization interval.
        beta_tol: Convergence tolerance for beta.
        noise_min: Lower bound for search.
        noise_max: Upper bound for search.

    Returns:
        Calibrated noise_multiplier (float).
    """
    # Handle backward compatibility for deprecated parameter names
    if beta_error is not None:
        warnings.warn(
            "Parameter 'beta_error' is deprecated and will be removed in v2.0.0. "
            "Use 'beta_tol' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        beta_tol = beta_error
    elif mu_error is not None:
        warnings.warn(
            "Parameter 'mu_error' is deprecated and will be removed in v2.0.0. "
            "Use 'beta_tol' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        beta_tol = mu_error
    elif beta_tol is None:
        beta_tol = 1e-3

    if mu_min is not None:
        warnings.warn(
            "Parameter 'mu_min' is deprecated and will be removed in v2.0.0. "
            "Use 'noise_min' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        noise_min = mu_min

    if mu_max is not None:
        warnings.warn(
            "Parameter 'mu_max' is deprecated and will be removed in v2.0.0. "
            "Use 'noise_max' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        noise_max = mu_max

    evaluator = create_dpsgd_evaluator(
        sample_rate=sample_rate,
        num_steps=num_steps,
        grid_step=grid_step,
        target_alpha=alpha,  # Pass alpha so evaluator computes beta
    )

    target = CalibrationTarget(kind="err_rates", alpha=alpha, beta=beta)

    config = CalibrationConfig(
        param_min=noise_min,
        param_max=noise_max,
        target_tol=beta_tol,
        increasing=True,
    )

    result = calibrate_parameter(
        evaluator, target, config, parameter_name="noise_multiplier"
    )

    return result.parameter_value


# Helper functions for analysis (not calibration)


def get_advantage_for_dpsgd(
    noise_multiplier: float, sample_rate: float, num_steps: int, grid_step: float = 1e-4
) -> float:
    """
    Compute advantage for DP-SGD with given parameters.

    Args:
        noise_multiplier: Noise scale parameter.
        sample_rate: Poisson sampling rate.
        num_steps: Number of DP-SGD steps.
        grid_step: PLD discretization interval.

    Returns:
        Attack advantage value.
    """
    evaluator = create_dpsgd_evaluator(sample_rate, num_steps, grid_step)
    metrics = evaluator(noise_multiplier)
    return metrics.advantage


def get_beta_for_dpsgd(
    noise_multiplier: float,
    sample_rate: float,
    num_steps: int,
    alpha: Union[float, np.ndarray],
    grid_step: float = 1e-4,
) -> Union[float, np.ndarray]:
    """
    Compute beta (FNR) for DP-SGD at given alpha (FPR).

    Args:
        noise_multiplier: Noise scale parameter.
        sample_rate: Poisson sampling rate.
        num_steps: Number of DP-SGD steps.
        alpha: False positive rate(s) in [0, 1]. Can be scalar or array.
        grid_step: PLD discretization interval.

    Returns:
        False negative rate(s) corresponding to input alpha.
    """
    from riskcal.analysis import get_beta_from_pld

    pld = pld_module.from_gaussian_mechanism(
        standard_deviation=noise_multiplier,
        sampling_prob=sample_rate,
        use_connect_dots=True,
        value_discretization_interval=grid_step,
    ).self_compose(num_steps)

    return get_beta_from_pld(pld, alpha=alpha)



import numpy as np
import pytest

from riskcal.calibration.dpsgd import (
    find_noise_multiplier_for_advantage,
    find_noise_multiplier_for_err_rates,
    get_advantage_for_dpsgd,
    get_beta_for_dpsgd,
)
from riskcal.accountants import CTDAccountant
from riskcal.calibration.blackbox import (
    find_noise_multiplier_for_advantage as blackbox_find_noise_multiplier_for_advantage,
    find_noise_multiplier_for_err_rates as blackbox_find_noise_multiplier_for_err_rates,
)
from scipy.stats import norm
from scipy.optimize import root_scalar


grid_step = 1e-2
sample_rate = 0.01
num_dpsgd_steps = 100


@pytest.fixture
def accountant():
    yield CTDAccountant


@pytest.mark.parametrize(
    "advantage, sample_rate, num_steps",
    [
        (0.01, 1, 1),
        (0.05, 1, 1),
        (0.10, 1, 1),
        (0.01, sample_rate, num_dpsgd_steps),
        (0.05, sample_rate, num_dpsgd_steps),
        (0.10, sample_rate, num_dpsgd_steps),
    ],
)
def test_advantage_calibration_correctness(advantage, sample_rate, num_steps):
    advantage_tol = 1e-2

    calibrated_mu = find_noise_multiplier_for_advantage(
        advantage=advantage,
        sample_rate=sample_rate,
        num_steps=num_steps,
        advantage_tol=advantage_tol,
        grid_step=grid_step,
    )

    numerical_advantage = get_advantage_for_dpsgd(
        noise_multiplier=calibrated_mu,
        sample_rate=sample_rate,
        num_steps=num_steps,
        grid_step=grid_step,
    )

    assert pytest.approx(numerical_advantage, abs=advantage_tol) == advantage


@pytest.mark.parametrize(
    "alpha, beta, sample_rate, num_steps",
    [
        # DP-SGD
        (0.1, 0.33, sample_rate, num_dpsgd_steps),
        (0.1, 0.50, sample_rate, num_dpsgd_steps),
        (0.1, 0.70, sample_rate, num_dpsgd_steps),
    ],
)
def test_err_rates_calibration_correctness(alpha, beta, sample_rate, num_steps):
    beta_tol = 1e-2
    calibrated_mu = find_noise_multiplier_for_err_rates(
        alpha=alpha,
        beta=beta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        beta_tol=beta_tol,
        grid_step=grid_step,
    )

    numerical_beta = get_beta_for_dpsgd(
        alpha=alpha,
        noise_multiplier=calibrated_mu,
        sample_rate=sample_rate,
        num_steps=num_steps,
        grid_step=grid_step,
    )

    assert pytest.approx(numerical_beta, abs=beta_tol) == beta


@pytest.mark.parametrize(
    "advantage, sample_rate, num_steps",
    [
        # DP-SGD
        (0.01, sample_rate, num_dpsgd_steps),
        (0.05, sample_rate, num_dpsgd_steps),
        (0.10, sample_rate, num_dpsgd_steps),
    ],
)
def test_advantage_calibration_blackbox_vs_direct(
    accountant, advantage, sample_rate, num_steps
):
    eps_error = 1e-2
    advantage_tol = 0.01
    mu_error = 0.01

    direct_calibrated_mu = find_noise_multiplier_for_advantage(
        advantage=advantage,
        sample_rate=sample_rate,
        num_steps=num_steps,
        advantage_tol=advantage_tol,
        grid_step=grid_step,
    )

    blackbox_calibrated_mu = blackbox_find_noise_multiplier_for_advantage(
        accountant=accountant,
        advantage=advantage,
        sample_rate=sample_rate,
        num_steps=num_steps,
        mu_error=mu_error,
        eps_error=eps_error,
        grid_step=grid_step,
    )

    assert direct_calibrated_mu == pytest.approx(blackbox_calibrated_mu, abs=mu_error)


# @pytest.mark.skip("Investigate big difference between direct and blackbox.")
@pytest.mark.parametrize(
    "alpha, beta, sample_rate, num_steps",
    [
        # DP-SGD
        (0.1, 0.33, sample_rate, num_dpsgd_steps),
        (0.1, 0.50, sample_rate, num_dpsgd_steps),
        (0.1, 0.70, sample_rate, num_dpsgd_steps),
    ],
)
def test_err_rates_calibration_blackbox_vs_direct(
    accountant, alpha, beta, sample_rate, num_steps
):
    """This test is very slow, since blackbox calibration using a PLD accountant is slow. So, we
    only test one (alpha,beta) point with generous errors. Noise parameter (i.e. mu) is roughly
    0.374 with these parameters.

    """
    eps_error = 1e-1
    beta_tol = 1e-1
    mu_error = 1e-1
    direct_calibrated_mu = find_noise_multiplier_for_err_rates(
        alpha=alpha,
        beta=beta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        beta_tol=beta_tol,
        grid_step=grid_step,
    )

    blackbox_calibration_result = blackbox_find_noise_multiplier_for_err_rates(
        accountant,
        alpha=alpha,
        beta=beta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        eps_error=eps_error,
        grid_step=grid_step,
    )
    blackbox_calibrated_mu = blackbox_calibration_result.noise_multiplier

    assert pytest.approx(blackbox_calibrated_mu, abs=mu_error) == direct_calibrated_mu



"""Privacy risk metrics and conversions.

This module provides functions for computing privacy metrics (beta, advantage,
Bayes risk) from different privacy representations (PLDs, ADP, GDP,
RDP, zCDP).
"""

from typing import Union, Optional
import warnings
import numpy as np
from scipy import stats, optimize
from dp_accounting.pld import privacy_loss_distribution

from riskcal.analysis import _plrv
from riskcal.utils import _ensure_array

from riskcal.analysis.rdp import get_FNR as _get_FNR_from_rdp


# =============================================================================
# Internal conversions
# =============================================================================


def pld_to_plrvs(
    pld: privacy_loss_distribution.PrivacyLossDistribution,
) -> _plrv.PLRVs:
    """
    Convert PLD to internal PLRVs representation.

    Converts a Google dp_accounting Privacy Loss Distribution object
    into the PLRVs (Privacy Loss Random Variables) format used internally.

    Args:
        pld: Privacy loss distribution from Google's dp_accounting library.

    Returns:
        PLRVs object containing the privacy loss random variables.

    Note:
        This is an internal conversion function. Most users should use
        the higher-level `get_beta_from_pld()`, `get_advantage_from_pld()`, etc.
    """

    def _get_plrv(pld):
        pld = pld.to_dense_pmf()
        pmf = pld._probs
        lower_loss = pld._lower_loss
        infinity_mass = pld._infinity_mass
        return lower_loss, infinity_mass, pmf

    lower_loss_Y, infinity_mass_Y, pmf_Y = _get_plrv(pld._pmf_remove)
    lower_loss_Z, infinity_mass_Z, pmf_Z = _get_plrv(pld._pmf_add)

    upper_loss_Z = lower_loss_Z + len(pmf_Z) - 1

    # clean pmfs. Sometimes float errors cause probs to be negative
    pmf_Y = np.where(pmf_Y < 0, 0, pmf_Y)
    pmf_Z = np.where(pmf_Z < 0, 0, pmf_Z)

    pmf_Y = pmf_Y * (1 - infinity_mass_Y) / np.sum(pmf_Y)
    pmf_Z = pmf_Z * (1 - infinity_mass_Z) / np.sum(pmf_Z)

    is_symmetric = pld._symmetric
    return _plrv.PLRVs(
        y0=lower_loss_Y,
        x0=-upper_loss_Z,
        pmf_Y=pmf_Y,
        pmf_X=pmf_Z[::-1],
        minus_infinity_mass_X=infinity_mass_Z,
        infinity_mass_Y=infinity_mass_Y,
        is_symmetric=is_symmetric,
    )


# Backward compatibility alias
plrvs_from_pld = pld_to_plrvs
plrvs_from_pld.__deprecated__ = "2.0.0"


# =============================================================================
# PLD (Privacy Loss Distribution)
# =============================================================================


def get_beta_from_pld(
    pld: privacy_loss_distribution.PrivacyLossDistribution,
    alpha: Union[float, np.ndarray] = None,
    alphas: Union[float, np.ndarray] = None,  # Deprecated
) -> Union[float, np.ndarray]:
    """
    Compute false negative rate (FNR) for given false positive rate (FPR) from PLD.

    Uses the direct method from Algorithm 1 (Kulynych et al., 2024) to compute
    the optimal trade-off between FNR (beta) and FPR (alpha).

    .. deprecated:: 1.2.0
        Parameter 'alphas' is deprecated and will be removed in version 2.0.0.
        Use 'alpha' instead.

    Args:
        pld: Privacy loss distribution from Google's dp_accounting library.
        alpha: False positive rate(s) in [0, 1]. Can be scalar or array.
        alphas: (Deprecated) Use 'alpha' instead.

    Returns:
        False negative rate(s) corresponding to input alpha.

    Example:
        >>> from dp_accounting.pld import privacy_loss_distribution as pld_module
        >>> pld = pld_module.from_gaussian_mechanism(1.0)
        >>> beta = get_beta_from_pld(pld, alpha=0.01)

    References:
        Kulynych & Gomez et al. (2024), Algorithm 1. https://arxiv.org/abs/2407.02191
    """
    if alpha is None and alphas is None:
        raise ValueError("Must specify alpha.")
    elif alpha is not None and alphas is not None:
        raise ValueError("Must pass either alpha or alphas.")
    elif alphas is not None:
        warnings.warn(
            "Parameter 'alphas' is deprecated. Use 'alpha' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        alpha = alphas

    return _plrv.get_beta(pld_to_plrvs(pld), alpha)


def get_advantage_from_pld(
    pld: privacy_loss_distribution.PrivacyLossDistribution,
) -> float:
    """
    Compute attack advantage from PLD.

    Advantage is the maximum value of (TPR - FPR) achievable by any attacker,
    which equals delta at epsilon=0.

    Args:
        pld: Privacy loss distribution from Google's dp_accounting library.

    Returns:
        Maximum attack advantage.

    References:
        Kulynych & Gomez et al. (2024). https://arxiv.org/abs/2407.02191
    """
    return pld.get_delta_for_epsilon(0)


def get_bayes_risk_from_pld(pld, prior):
    """
    Compute Bayes Risk from PLD for given prior probability.

    Bayes risk shows the maximum accuracy of an attack against privacy of a single
    record under a binary prior (e.g., accuracy of attribute inference).

    Args:
        pld: Privacy loss distribution from Google's dp_accounting library.
        prior: Prior probability (scalar or array). Probability that the
            sensitive attribute takes value 1.

    Returns:
        Bayes risk value(s). Float if prior is scalar, array if prior is array.

    Example:
        >>> from dp_accounting.pld import privacy_loss_distribution as pld_module
        >>> pld = pld_module.from_laplace_mechanism(1.0)
        >>> risk = get_bayes_risk_from_pld(pld, prior=0.5)

    References:
        Kulynych et al. (2025), Proposition D.1 / Eq. 35.
        https://arxiv.org/abs/2507.06969
    """
    prior, is_scalar = _ensure_array(prior)
    epsilon = np.log(prior / (1 - prior))
    delta = np.array([pld.get_delta_for_epsilon(e) for e in epsilon])
    bayes_risk = (1 - prior) * (1 - delta)

    if is_scalar:
        return bayes_risk[0]
    return np.array(bayes_risk)


# =============================================================================
# GDP (Gaussian Differential Privacy)
# =============================================================================


def get_beta_from_gdp(
    mu: float, alpha: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Compute FNR for FPR using the analytical formula for Gaussian DP.

    Args:
        mu: Gaussian noise scale parameter (sigma).
        alpha: False positive rate(s) in [0, 1].

    Returns:
        False negative rate(s) corresponding to input alpha.

    References:
        Dong et al. (2019), Eq. 6. https://arxiv.org/abs/1905.02383
    """
    return stats.norm.cdf(-stats.norm.ppf(alpha) - mu)


def get_advantage_from_gdp(mu: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    Compute attack advantage using analytical formula for Gaussian mechanism.

    Args:
        mu: Gaussian noise scale parameter (sigma). Can be scalar or array.

    Returns:
        Attack advantage value(s).

    References:
        Dong et al. (2019), Corollary 2.13. https://arxiv.org/abs/1905.02383
    """
    return stats.norm.cdf(mu / 2) - stats.norm.cdf(-mu / 2)


def get_bayes_risk_from_gdp(mu, prior):
    """
    Compute Bayes Risk for Gaussian mechanism using analytical formula.

    Args:
        mu: Gaussian noise scale parameter (sigma).
        prior: Prior probability (scalar or array).

    Returns:
        Bayes risk value(s). Float if prior is scalar, array if prior is array.

    References:
        Kulynych et al. (2025), Proposition D.1 / Eq. 35.
        https://arxiv.org/abs/2507.06969
        Dong et al. (2019), Corollary 2.13. https://arxiv.org/abs/1905.02383
    """
    assert mu >= 0, "mu must be >= 0"
    prior, is_scalar = _ensure_array(prior)
    epsilon = np.log(prior / (1 - prior))
    # Gaussian privacy profile: Dong et al. (2019), Corollary 2.13
    delta = stats.norm.cdf(-epsilon / mu + mu / 2) - np.exp(epsilon) * stats.norm.cdf(
        -epsilon / mu - mu / 2
    )
    bayes_risk = (1 - prior) * (1 - delta)

    if is_scalar:
        return bayes_risk[0]
    return np.array(bayes_risk)


# Backward compatibility aliases
get_beta_for_mu = get_beta_from_gdp
get_beta_for_mu.__deprecated__ = "2.0.0"
get_advantage_for_mu = get_advantage_from_gdp
get_advantage_for_mu.__deprecated__ = "2.0.0"
get_bayes_risk_for_mu = get_bayes_risk_from_gdp
get_bayes_risk_for_mu.__deprecated__ = "2.0.0"


# =============================================================================
# ADP (Approximate Differential Privacy)
# =============================================================================


def get_beta_from_adp(
    epsilon: float, delta: float, alpha: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Compute FNR for FPR from (epsilon, delta)-DP parameters.

    Args:
        epsilon: Privacy parameter epsilon.
        delta: Privacy parameter delta.
        alpha: False positive rate(s) in [0, 1].

    Returns:
        False negative rate(s) corresponding to input alpha.

    Example:
        >>> import numpy as np
        >>> np.round(get_beta_from_adp(1.0, 0.001, 0.8), 3)
        0.073

    References:
        Dong et al. (2019), Eq. 5. https://arxiv.org/abs/1905.02383
    """
    form1 = np.array(1 - delta - np.exp(epsilon) * alpha)
    form2 = np.array(np.exp(-epsilon) * (1 - delta - alpha))
    return np.maximum.reduce([form1, form2, np.zeros_like(form1)])


def get_advantage_from_adp(epsilon: float, delta: float) -> float:
    """
    Compute attack advantage from (epsilon, delta)-DP parameters.

    Args:
        epsilon: Privacy parameter epsilon.
        delta: Privacy parameter delta.

    Returns:
        Attack advantage.

    Example:
        >>> import numpy as np
        >>> np.round(get_advantage_from_adp(0., 0.001), 3)
        0.001

    References:
        Dong et al. (2019). https://arxiv.org/abs/1905.02383
    """
    return (np.exp(epsilon) + 2 * delta - 1) / (np.exp(epsilon) + 1)


def get_epsilon_from_err_rates(delta: float, alpha: float, beta: float) -> float:
    """
    Convert f-DP error rates (alpha, beta) to epsilon for (epsilon, delta)-DP.

    Args:
        delta: Target delta parameter for (epsilon, delta)-DP.
        alpha: False positive rate (FPR) from f-DP.
        beta: False negative rate (FNR) from f-DP.

    Returns:
        Epsilon value corresponding to the error rates.

    Example:
        >>> import numpy as np
        >>> np.round(get_epsilon_from_err_rates(0.001, 0.001, 0.8), 3)
        5.293

    References:
        Dong et al. (2019). https://arxiv.org/abs/1905.02383
    """
    epsilon1 = np.log((1 - delta - alpha) / beta)
    epsilon2 = np.log((1 - delta - beta) / alpha)
    return np.maximum.reduce([epsilon1, epsilon2, np.zeros_like(epsilon1)])


# Backward compatibility aliases
get_beta_for_epsilon_delta = get_beta_from_adp
get_beta_for_epsilon_delta.__deprecated__ = "2.0.0"
get_advantage_for_epsilon_delta = get_advantage_from_adp
get_advantage_for_epsilon_delta.__deprecated__ = "2.0.0"
get_epsilon_for_err_rates = get_epsilon_from_err_rates
get_epsilon_for_err_rates.__deprecated__ = "2.0.0"


# =============================================================================
# RDP (Renyi Differential Privacy)
# =============================================================================


def get_beta_from_rdp(
    epsilon: float,
    alpha: Union[float, np.ndarray],
    order: float,
    linear_search_step: float = 1e-3,  # unused; kept for backward compatibility
    max_bisection_steps: int = 50,  # unused; kept for backward compatibility
    tol: float = 1e-7,
) -> Union[float, np.ndarray]:
    """
    Compute FNR for FPR from Renyi DP parameters.

    Uses the optimal conversion from a single RDP guarantee to a tradeoff
    function (Riess et al., 2026). Accepts scalar or array FPR values.

    Args:
        epsilon: Renyi divergence parameter (privacy budget).
        alpha: False positive rate(s) (FPR) in [0, 1]. Scalar or array.
        order: Order of Renyi divergence (alpha in Renyi DP literature).
        linear_search_step: Unused. Kept for backward compatibility.
        max_bisection_steps: Unused. Kept for backward compatibility.
        tol: Numerical tolerance for bisection convergence.

    Returns:
        False negative rate (FNR) corresponding to input alpha.

    References:
        Zhu et al. (2022), Appendix F.1. https://arxiv.org/abs/2106.08567
        Riess et al. (2026). https://arxiv.org/abs/2602.04562
    """
    return _get_FNR_from_rdp(alpha_array=alpha, order=order, epsilon=epsilon, tol=tol)


# =============================================================================
# zCDP (Zero-Concentrated Differential Privacy)
# =============================================================================


def get_beta_from_zcdp(
    rho: float,
    alpha: Union[float, np.ndarray],
    tol: float = 1e-9,
    min_order: float = 1.0,
    max_order: float = 1024.0,
    grid_size=1000,
) -> Union[float, np.ndarray]:
    """
    Compute FNR for FPR from zCDP parameter rho.

    Zero-Concentrated Differential Privacy (zCDP) is characterized by a single
    parameter rho. This function computes the optimal trade-off between false
    negative rate (beta) and false positive rate (alpha) by optimizing over
    Renyi divergence orders.

    Args:
        rho: Zero-concentrated differential privacy parameter.
        alpha: False positive rate(s) in [0, 1]. Can be scalar or array.
        tol: Tolerance for numerical convergence and avoiding log(0).
        max_order: Maximum Renyi DP order to consider.

    Returns:
        False negative rate(s) corresponding to input alpha. Returns float if
        alpha is scalar, array if alpha is array.

    Example:
        >>> import numpy as np
        >>> # Single alpha value
        >>> beta = get_beta_from_zcdp(rho=0.5, alpha=0.1)
        >>> np.round(beta, 3)
        0.517
        >>> # Multiple alpha values
        >>> betas = get_beta_from_zcdp(rho=0.5, alpha=np.array([0.1, 0.2, 0.3]))
        >>> np.round(betas, 3)
        array([0.517, 0.34, 0.232])

    References:
        Bun & Steinke (2016). https://arxiv.org/abs/1605.02065
        Zhu et al. (2022), Appendix F.1. https://arxiv.org/abs/2106.08567
        Riess et al. (2026). https://arxiv.org/abs/2602.04562
    """
    scalar_input = isinstance(alpha, (int, float))
    alpha_arr = np.atleast_1d(np.asarray(alpha, dtype=float))

    result = np.empty_like(alpha_arr)
    low = alpha_arr <= tol
    high = alpha_arr >= 1 - tol
    interior = ~low & ~high

    result[low] = 1.0
    result[high] = 0.0

    orders_grid = np.concatenate(
        [
            np.linspace(min_order, 2, grid_size // 3)[:-1],
            np.linspace(2, 16, grid_size // 3)[:-1],
            np.logspace(np.log10(16), np.log10(max_order), grid_size // 3),
        ]
    )
    if interior.any():
        alpha_interior = alpha_arr[interior]
        best_vals = np.zeros(alpha_interior.shape)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            for order in orders_grid:
                best_vals = np.maximum(
                    best_vals,
                    np.atleast_1d(
                        get_beta_from_rdp(
                            epsilon=rho * order,
                            alpha=alpha_interior,
                            order=order,
                            tol=tol,
                        )
                    ),
                )
        result[interior] = best_vals

    return float(result[0]) if scalar_input else result


def get_advantage_from_zcdp(
    rho: float,
    tol: float = 1e-9,
    min_order: float = 1.0,
    max_order: float = 1024.0,
    grid_size: int = 1000,
) -> float:
    """
    Compute attack advantage from zCDP parameter rho.

    Advantage is the maximum value of (TPR - FPR) achievable by any attacker.
    This function optimizes over all possible threshold choices to find the
    maximum advantage.

    Args:
        rho: Zero-concentrated differential privacy parameter.
        tol: Tolerance for numerical convergence and avoiding log(0).
        min_order: Minimum Renyi DP order to consider.
        max_order: Maximum Renyi DP order to consider.
        grid_size: Number of orders to evaluate in the grid search.

    Returns:
        Maximum attack advantage.

    Example:
        >>> import numpy as np
        >>> adv = get_advantage_from_zcdp(rho=0.5)
        >>> np.round(adv, 3)
        0.47

    References:
        Bun & Steinke (2016). https://arxiv.org/abs/1605.02065
    """
    result = optimize.minimize_scalar(
        lambda alpha: -(
            1 - alpha - get_beta_from_zcdp(
                rho=rho, alpha=alpha, tol=tol, min_order=min_order,
                max_order=max_order, grid_size=grid_size,
            )
        ),
        bounds=(0, 1),
        method="bounded",
        options={"xatol": 1e-12},
    )
    if not result.success:
        warnings.warn("Optimization failed for advantage calculation")
        return np.nan
    else:
        return -result.fun


def get_mu_from_zcdp_approx(rho: float) -> float:
    """Deprecated. Use get_beta_from_zcdp or get_advantage_from_zcdp instead."""
    warnings.warn(
        "get_mu_from_zcdp_approx is deprecated and will be removed in a future version. "
        "Use get_beta_from_zcdp or get_advantage_from_zcdp for exact computations.",
        DeprecationWarning,
        stacklevel=2,
    )
    return (
        1.5822558654881096 * np.sqrt(rho)
        + 0.08064620797681155 * rho
        + 0.05485600531538526
    )


def get_beta_from_zcdp_approx(
    rho: float, alpha: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """Deprecated. Use get_beta_from_zcdp instead."""
    warnings.warn(
        "get_beta_from_zcdp_approx is deprecated and will be removed in a future version. "
        "Use get_beta_from_zcdp for exact computation.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_beta_from_zcdp(rho, alpha=alpha)


def get_advantage_from_zcdp_approx(rho: float) -> float:
    """Deprecated. Use get_advantage_from_zcdp instead."""
    warnings.warn(
        "get_advantage_from_zcdp_approx is deprecated and will be removed in a future version. "
        "Use get_advantage_from_zcdp for exact computation.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_advantage_from_zcdp(rho)



"""Tests for backward compatibility and deprecation warnings."""

import pytest
import numpy as np
import sys
import importlib
from dp_accounting.pld.privacy_loss_distribution import from_gaussian_mechanism


# Global tolerances for test assertions
REL_TOLERANCE = 1e-4
ABS_TOLERANCE = 1e-12


# ============================================================================
# Deprecated Module Imports
# ============================================================================


def test_dpsgd_module_import_warns():
    """Test that importing riskcal.dpsgd emits deprecation warning."""
    # Remove from sys.modules if already imported
    if "riskcal.dpsgd" in sys.modules:
        del sys.modules["riskcal.dpsgd"]

    with pytest.warns(DeprecationWarning, match="riskcal.dpsgd.*deprecated"):
        import riskcal.dpsgd  # noqa: F401


def test_conversions_module_import_warns():
    """Test that importing riskcal.conversions emits deprecation warning."""
    # Remove from sys.modules if already imported
    if "riskcal.conversions" in sys.modules:
        del sys.modules["riskcal.conversions"]

    with pytest.warns(DeprecationWarning, match="riskcal.conversions.*deprecated"):
        import riskcal.conversions  # noqa: F401


def test_blackbox_module_import_warns():
    """Test that importing riskcal.blackbox emits deprecation warning."""
    # Remove from sys.modules if already imported
    if "riskcal.blackbox" in sys.modules:
        del sys.modules["riskcal.blackbox"]

    with pytest.warns(DeprecationWarning, match="riskcal.blackbox.*deprecated"):
        import riskcal.blackbox  # noqa: F401


def test_plrv_module_import_warns():
    """Test that importing riskcal.plrv emits deprecation warning."""
    # Remove from sys.modules if already imported
    if "riskcal.plrv" in sys.modules:
        del sys.modules["riskcal.plrv"]

    with pytest.warns(DeprecationWarning, match="riskcal.plrv.*deprecated"):
        import riskcal.plrv  # noqa: F401


# ============================================================================
# Deprecated Parameter Names
# ============================================================================


def test_get_beta_from_pld_alphas_parameter_warns():
    """Test that 'alphas' parameter emits deprecation warning."""
    pld = from_gaussian_mechanism(1.0, value_discretization_interval=1e-3)

    # Import from new location
    from riskcal.analysis import get_beta_from_pld

    with pytest.warns(DeprecationWarning, match="'alphas'.*deprecated"):
        beta = get_beta_from_pld(pld, alphas=0.1)
        assert isinstance(beta, (float, np.floating))


def test_find_noise_multiplier_advantage_error_param_warns():
    """Test that deprecated 'advantage_error' parameter warns."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_advantage

    with pytest.warns(DeprecationWarning, match="'advantage_error'.*deprecated"):
        noise = find_noise_multiplier_for_advantage(
            advantage=0.1,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-2,
            advantage_error=1e-2,
        )
        assert noise > 0


def test_find_noise_multiplier_mu_min_param_warns():
    """Test that deprecated 'mu_min' parameter warns."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_advantage

    with pytest.warns(DeprecationWarning, match="'mu_min'.*deprecated"):
        noise = find_noise_multiplier_for_advantage(
            advantage=0.5,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-2,
            mu_min=0.1,
        )
        assert noise > 0


def test_find_noise_multiplier_mu_max_param_warns():
    """Test that deprecated 'mu_max' parameter warns."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_advantage

    with pytest.warns(DeprecationWarning, match="'mu_max'.*deprecated"):
        noise = find_noise_multiplier_for_advantage(
            advantage=0.5,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-2,
            mu_max=50.0,
        )
        assert noise > 0


def test_find_noise_multiplier_mu_error_param_warns():
    """Test that deprecated 'mu_error' parameter warns."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_advantage

    with pytest.warns(DeprecationWarning, match="'mu_error'.*deprecated"):
        noise = find_noise_multiplier_for_advantage(
            advantage=0.5,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-1,
            mu_error=1e-2,
        )
        assert noise > 0


def test_find_noise_multiplier_err_rates_beta_error_param_warns():
    """Test that deprecated 'beta_error' parameter warns."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_err_rates

    with pytest.warns(DeprecationWarning, match="'beta_error'.*deprecated"):
        noise = find_noise_multiplier_for_err_rates(
            alpha=0.1,
            beta=0.2,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-1,
            beta_error=1e-2,
        )
        assert noise > 0


def test_find_noise_multiplier_err_rates_mu_min_param_warns():
    """Test that deprecated 'mu_min' parameter warns in err_rates function."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_err_rates

    with pytest.warns(DeprecationWarning, match="'mu_min'.*deprecated"):
        noise = find_noise_multiplier_for_err_rates(
            alpha=0.1,
            beta=0.2,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-1,
            mu_min=0.1,
        )
        assert noise > 0


def test_find_noise_multiplier_err_rates_mu_max_param_warns():
    """Test that deprecated 'mu_max' parameter warns in err_rates function."""
    from riskcal.calibration.dpsgd import find_noise_multiplier_for_err_rates

    with pytest.warns(DeprecationWarning, match="'mu_max'.*deprecated"):
        noise = find_noise_multiplier_for_err_rates(
            alpha=0.1,
            beta=0.2,
            sample_rate=1,
            num_steps=1,
            grid_step=1e-1,
            mu_max=50.0,
        )
        assert noise > 0


# ============================================================================
# Deprecated Function Aliases
# ============================================================================


def test_get_beta_for_mu_alias_works():
    """Test that get_beta_for_mu is an alias for get_beta_from_gdp."""
    from riskcal.analysis import get_beta_for_mu, get_beta_from_gdp

    noise_multiplier = 1.0
    alpha = 0.1

    beta_alias = get_beta_for_mu(noise_multiplier, alpha)
    beta_new = get_beta_from_gdp(noise_multiplier, alpha)

    assert beta_alias == pytest.approx(beta_new, rel=REL_TOLERANCE, abs=ABS_TOLERANCE)


def test_get_advantage_for_mu_alias_works():
    """Test that get_advantage_for_mu is an alias for get_advantage_from_gdp."""
    from riskcal.analysis import get_advantage_for_mu, get_advantage_from_gdp

    noise_multiplier = 1.0

    advantage_alias = get_advantage_for_mu(noise_multiplier)
    advantage_new = get_advantage_from_gdp(noise_multiplier)

    assert advantage_alias == pytest.approx(
        advantage_new, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


def test_get_bayes_risk_for_mu_alias_works():
    """Test that get_bayes_risk_for_mu is an alias for get_bayes_risk_from_gdp."""
    from riskcal.analysis import get_bayes_risk_for_mu, get_bayes_risk_from_gdp

    noise_multiplier = 1.0
    prior = 0.5

    risk_alias = get_bayes_risk_for_mu(noise_multiplier, prior)
    risk_new = get_bayes_risk_from_gdp(noise_multiplier, prior)

    assert risk_alias == pytest.approx(risk_new, rel=REL_TOLERANCE, abs=ABS_TOLERANCE)


def test_plrvs_from_pld_alias_works():
    """Test that plrvs_from_pld is an alias for pld_to_plrvs."""
    from riskcal.analysis import plrvs_from_pld, pld_to_plrvs

    pld = from_gaussian_mechanism(1.0, value_discretization_interval=1e-3)

    plrvs_alias = plrvs_from_pld(pld)
    plrvs_new = pld_to_plrvs(pld)

    # Check that both produce the same structure
    assert plrvs_alias.y0 == plrvs_new.y0
    assert plrvs_alias.x0 == plrvs_new.x0
    assert np.allclose(plrvs_alias.pmf_Y, plrvs_new.pmf_Y)
    assert np.allclose(plrvs_alias.pmf_X, plrvs_new.pmf_X)


def test_get_epsilon_for_err_rates_alias_works():
    """Test that get_epsilon_for_err_rates is an alias for get_epsilon_from_err_rates."""
    from riskcal.analysis import get_epsilon_for_err_rates, get_epsilon_from_err_rates

    delta = 0.001
    alpha = 0.01
    beta = 0.1

    epsilon_alias = get_epsilon_for_err_rates(delta, alpha, beta)
    epsilon_new = get_epsilon_from_err_rates(delta, alpha, beta)

    assert epsilon_alias == pytest.approx(
        epsilon_new, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


# ============================================================================
# Legacy API Still Works
# ============================================================================


def test_legacy_dpsgd_get_advantage_for_dpsgd_works():
    """Test that get_advantage_for_dpsgd from deprecated module works."""
    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.dpsgd import get_advantage_for_dpsgd

    advantage = get_advantage_for_dpsgd(
        noise_multiplier=1.0, sample_rate=0.01, num_steps=100
    )
    assert 0 < advantage < 1


def test_legacy_dpsgd_get_beta_for_dpsgd_works():
    """Test that get_beta_for_dpsgd from deprecated module works."""
    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.dpsgd import get_beta_for_dpsgd

    beta = get_beta_for_dpsgd(
        noise_multiplier=1.0, sample_rate=0.01, num_steps=100, alpha=0.1
    )
    assert 0 < beta < 1


def test_legacy_dpsgd_find_noise_multiplier_for_advantage_works():
    """Test that find_noise_multiplier_for_advantage from deprecated module works."""
    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.dpsgd import find_noise_multiplier_for_advantage

    noise = find_noise_multiplier_for_advantage(
        advantage=0.5, sample_rate=1, num_steps=1, grid_step=1e-2
    )
    assert noise > 0


def test_legacy_conversions_get_beta_from_pld_works():
    """Test that get_beta_from_pld from deprecated module works."""
    pld = from_gaussian_mechanism(1.0, value_discretization_interval=1e-3)

    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.conversions import get_beta_from_pld

    beta = get_beta_from_pld(pld, alpha=0.1)
    assert 0 < beta < 1


def test_legacy_conversions_get_advantage_from_pld_works():
    """Test that get_advantage_from_pld from deprecated module works."""
    pld = from_gaussian_mechanism(1.0, value_discretization_interval=1e-3)

    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.conversions import get_advantage_from_pld

    advantage = get_advantage_from_pld(pld)
    assert 0 < advantage < 1


def test_legacy_CTDAccountant_works():
    """Test that CTDAccountant imported from deprecated module works."""
    # Import emits warning (may already be imported, so we don't check for warning)
    from riskcal.dpsgd import CTDAccountant

    acct = CTDAccountant()
    for _ in range(10):
        acct.step(noise_multiplier=1.0, sample_rate=0.01)

    beta = acct.get_beta(alpha=0.1)
    assert 0 < beta < 1

    advantage = acct.get_advantage()
    assert 0 < advantage < 1


# ============================================================================
# CalibrationResult Backward Compatibility
# ============================================================================


def test_calibration_result_noise_multiplier_property():
    """Test that CalibrationResult.noise_multiplier property works."""
    from riskcal.calibration.dpsgd import create_dpsgd_evaluator
    from riskcal.calibration.core import (
        CalibrationTarget,
        CalibrationConfig,
        calibrate_parameter,
    )

    evaluator = create_dpsgd_evaluator(sample_rate=1, num_steps=1)
    target = CalibrationTarget(kind="advantage", advantage=0.5)
    config = CalibrationConfig(param_min=0.05, param_max=10.0, increasing=False)

    result = calibrate_parameter(
        evaluator, target, config, parameter_name="noise_multiplier"
    )

    # Test backward compatibility property
    assert result.noise_multiplier == result.parameter_value
    assert result.noise_multiplier > 0



"""Utility functions for riskcal."""

from typing import Callable, Tuple
import numpy as np


def _ensure_array(x):
    """
    Convert input to numpy array and track if it was originally a scalar.

    Args:
        x: Input value (scalar or array-like)

    Returns:
        Tuple of (np.ndarray, bool) where bool indicates if input was scalar
    """
    is_scalar = isinstance(x, (int, float))
    if is_scalar:
        return np.asarray([x]), is_scalar
    return np.asarray(x), False


def inverse_monotone_function(
    f: Callable[[float], float],
    f_target: float,
    bounds: Tuple[float, float],
    func_threshold: float = 1e-3,
    param_threshold: float = 1e-3,
    max_iter: int = 100,
    increasing: bool = False,
) -> float:
    """
    Find the value of x such that the monotonic function f(x) is approximately
    equal to f_target within a given threshold using binary search.

    Args:
        f: A monotonic function (increasing or decreasing) to invert.
        f_target: The target value for f(x).
        bounds: A tuple (lower_x, upper_x) defining the search interval.
        func_threshold: Acceptable error for |f(x) - f_target|.
        increasing: Indicates if f is increasing (True) or decreasing (False).

    Returns:
        The value of x that satisfies the threshold conditions.

    Notes:
        It is guaranteed that the returned x is within the thresholds of the
        smallest (for monotonically decreasing func) or the largest (for
        monotonically increasing func) such x.
    """

    # Initialize bounds and midpoint
    lower_x, upper_x = bounds
    mid_x = (upper_x + lower_x) / 2
    f_mid = f(mid_x)

    # setup check function
    if increasing:
        check = lambda f_value, target_value: f_value <= target_value

        def continue_condition(upper_x, lower_x):
            return (abs(f(lower_x) - f_target) > func_threshold) or (
                abs(lower_x - upper_x) > param_threshold
            )

    else:
        check = lambda f_value, target_value: f_value > target_value

        def continue_condition(upper_x, lower_x):
            return (abs(f(upper_x) - f_target) > func_threshold) or (
                abs(lower_x - upper_x) > param_threshold
            )

    # run bisection
    for i in range(max_iter):

        if not continue_condition(upper_x, lower_x):
            break

        mid_x = (upper_x + lower_x) / 2
        f_mid = f(mid_x)
        if check(f_mid, f_target):
            lower_x = mid_x
        else:
            upper_x = mid_x

    if continue_condition(upper_x, lower_x):
        raise RuntimeError(
            f"Could not achieve the desired tolerance. Consider increasing "
            f"tolerance or increasing the parameter range. "
            f"{lower_x=} {upper_x=} / "
            f"param_diff={abs(upper_x - lower_x)} / target_diff={f(mid_x) - f_target}"
        )

    if increasing:
        return lower_x
    else:
        return upper_x



"""Generic privacy parameter calibration algorithm.

This module provides a unified calibration interface that works with any
privacy mechanism, provided you can supply an evaluator function that computes
privacy metrics for a given parameter value.
"""

from typing import Protocol, Literal, Any
from dataclasses import dataclass
import numpy as np

from riskcal.utils import inverse_monotone_function


class PrivacyEvaluator(Protocol):
    """Protocol for functions that evaluate privacy metrics for given parameter value."""

    def __call__(self, parameter_value: float) -> "PrivacyMetrics":
        """
        Evaluate privacy metrics for given parameter value.

        Args:
            parameter_value: Value of the parameter being calibrated
                (e.g., noise_multiplier, epsilon, sample_rate)

        Returns:
            PrivacyMetrics containing computed metrics
        """
        ...


@dataclass
class PrivacyMetrics:
    """Privacy metrics computed for a given parameter value."""

    # f-DP metrics
    advantage: float | None = None
    alpha: float | None = None  # FPR (if evaluating specific alpha)
    beta: float | None = None  # FNR (corresponding to alpha)

    # (ε,δ)-DP metrics
    epsilon: float | None = None
    delta: float | None = None

    # Additional metadata
    metadata: dict | None = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}


@dataclass
class CalibrationTarget:
    """Specification of target privacy level to calibrate to."""

    kind: Literal["advantage", "err_rates", "epsilon_delta"]

    # For advantage target
    advantage: float | None = None

    # For err_rates target
    alpha: float | None = None  # Target FPR
    beta: float | None = None  # Target FNR

    # For epsilon_delta target
    epsilon: float | None = None
    delta: float | None = None


@dataclass
class CalibrationConfig:
    """Configuration for calibration search."""

    increasing: bool = True
    param_min: float = 0.0  # Lower bound for parameter search
    param_max: float = 100.0  # Upper bound for parameter search
    target_tol: float = 1e-3  # Convergence tolerance for target metric
    param_tol: float = 1e-3  # Convergence tolerance for parameter
    max_iterations: int = 100


@dataclass
class CalibrationResult:
    """Result of parameter calibration."""

    parameter_value: float  # Calibrated parameter value
    parameter_name: str = "noise_multiplier"  # Name of calibrated parameter

    # Achieved metrics
    achieved_advantage: float | None = None
    achieved_alpha: float | None = None
    achieved_beta: float | None = None
    achieved_epsilon: float | None = None
    achieved_delta: float | None = None

    # Convergence info
    converged: bool = True
    iterations: int = 0
    method: str = "generic"

    # Additional info
    metadata: dict | None = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}

    # Backward compatibility property
    @property
    def noise_multiplier(self) -> float:
        """Backward compatibility: return parameter_value as noise_multiplier."""
        return self.parameter_value


def calibrate_parameter(
    evaluator: PrivacyEvaluator,
    target: CalibrationTarget,
    config: CalibrationConfig | None = None,
    parameter_name: str = "noise_multiplier",
) -> CalibrationResult:
    """
    Generic privacy parameter calibration algorithm.

    Finds the parameter value such that the privacy guarantee (as measured
    by the evaluator) meets or exceeds the specified target.

    This is the core calibration algorithm that works with any privacy mechanism,
    provided you can supply an evaluator function that computes privacy metrics
    for a given parameter value.

    Args:
        evaluator: Callable that maps parameter_value → PrivacyMetrics.
            Should compute the relevant privacy metrics (advantage, beta, epsilon)
            for a given parameter value.
        target: Target privacy level to calibrate to. Specifies what metric
            to optimize (advantage, err_rates, or epsilon_delta) and the target value.
        config: Calibration configuration (search bounds, tolerances, etc.).
            If None, uses default CalibrationConfig().
        parameter_name: Name of the parameter being calibrated (for documentation).
            Default: "noise_multiplier"

    Returns:
        CalibrationResult containing:
            - parameter_value: Calibrated parameter value
            - parameter_name: Name of the calibrated parameter
            - achieved_*: Actually achieved privacy metrics at this parameter value
            - converged: Whether calibration converged
            - iterations: Number of search iterations

    Raises:
        ValueError: If target specification is invalid or inconsistent
        RuntimeError: If calibration fails to converge

    Example:
        >>> # Define evaluator for DP-SGD
        >>> def evaluate_dpsgd(noise_mult):
        ...     pld = create_dpsgd_pld(noise_mult, sample_rate=0.002, num_steps=10000)
        ...     advantage = get_advantage_from_pld(pld)
        ...     return PrivacyMetrics(advantage=advantage)
        >>>
        >>> # Calibrate to advantage target
        >>> target = CalibrationTarget(kind='advantage', advantage=0.1)
        >>> result = calibrate_parameter(evaluate_dpsgd, target)
        >>> print(f"Noise: {result.parameter_value:.3f}")

    Notes:
        - Uses binary search for monotonic objectives (advantage, epsilon, beta)
        - Assumes higher parameter values → stronger privacy (lower advantage/epsilon/beta)
          For parameters where lower is better, adjust bounds accordingly
    """
    config = config or CalibrationConfig()

    # Validate target
    _validate_target(target)

    # Dispatch based on target type
    if target.kind == "advantage":
        return _calibrate_to_advantage(evaluator, target, config, parameter_name)
    elif target.kind == "err_rates":
        return _calibrate_to_err_rates(evaluator, target, config, parameter_name)
    elif target.kind == "epsilon_delta":
        return _calibrate_to_epsilon_delta(evaluator, target, config, parameter_name)
    else:
        raise ValueError(f"Unknown target kind: {target.kind}")


def _calibrate_to_advantage(
    evaluator: PrivacyEvaluator,
    target: CalibrationTarget,
    config: CalibrationConfig,
    parameter_name: str,
) -> CalibrationResult:
    """Calibrate to target advantage using binary search."""

    def objective(param_value: float) -> float:
        """Compute advantage for given parameter value."""
        metrics = evaluator(param_value)
        if metrics.advantage is None:
            raise ValueError("Evaluator must return advantage for advantage target")
        return metrics.advantage

    # Binary search: higher parameter → lower advantage (decreasing function)
    parameter_value = inverse_monotone_function(
        f=objective,
        f_target=target.advantage,
        bounds=(config.param_min, config.param_max),
        func_threshold=config.target_tol,
        max_iter=config.max_iterations,
        increasing=config.increasing,
    )

    # Evaluate final metrics
    final_metrics = evaluator(parameter_value)

    return CalibrationResult(
        parameter_value=parameter_value,
        parameter_name=parameter_name,
        achieved_advantage=final_metrics.advantage,
        achieved_epsilon=final_metrics.epsilon,
        achieved_delta=final_metrics.delta,
        converged=True,
        method="binary_search",
    )


def _calibrate_to_err_rates(
    evaluator: PrivacyEvaluator,
    target: CalibrationTarget,
    config: CalibrationConfig,
    parameter_name: str,
) -> CalibrationResult:
    """Calibrate to target (alpha, beta) using binary search on beta."""

    def objective(param_value: float) -> float:
        """Compute beta at target alpha for given parameter value."""
        metrics = evaluator(param_value)
        if metrics.beta is None:
            raise ValueError("Evaluator must return beta for err_rates target")
        return metrics.beta

    # Binary search: higher parameter → lower beta (decreasing function)
    parameter_value = inverse_monotone_function(
        f=objective,
        f_target=target.beta,
        bounds=(config.param_min, config.param_max),
        func_threshold=config.target_tol,
        max_iter=config.max_iterations,
        increasing=config.increasing,
    )

    # Evaluate final metrics
    final_metrics = evaluator(parameter_value)

    return CalibrationResult(
        parameter_value=parameter_value,
        parameter_name=parameter_name,
        achieved_alpha=target.alpha,
        achieved_beta=final_metrics.beta,
        achieved_advantage=final_metrics.advantage,
        converged=True,
        method="binary_search",
    )


def _calibrate_to_epsilon_delta(
    evaluator: PrivacyEvaluator,
    target: CalibrationTarget,
    config: CalibrationConfig,
    parameter_name: str,
) -> CalibrationResult:
    """Calibrate to target (epsilon, delta) using binary search on epsilon."""

    def objective(param_value: float) -> float:
        """Compute epsilon at target delta for given parameter value."""
        metrics = evaluator(param_value)
        if metrics.epsilon is None:
            raise ValueError("Evaluator must return epsilon for epsilon_delta target")
        return metrics.epsilon

    # Binary search: higher parameter → lower epsilon (decreasing function)
    parameter_value = inverse_monotone_function(
        f=objective,
        f_target=target.epsilon,
        bounds=(config.param_min, config.param_max),
        func_threshold=config.target_tol,
        max_iter=config.max_iterations,
        increasing=config.increasing,
    )

    # Evaluate final metrics
    final_metrics = evaluator(parameter_value)

    return CalibrationResult(
        parameter_value=parameter_value,
        parameter_name=parameter_name,
        achieved_epsilon=final_metrics.epsilon,
        achieved_delta=target.delta,
        converged=True,
        method="binary_search",
    )


def _validate_target(target: CalibrationTarget) -> None:
    """Validate target specification."""
    if target.kind == "advantage":
        if target.advantage is None:
            raise ValueError("advantage target requires advantage value")
        if not 0 <= target.advantage <= 1:
            raise ValueError(f"advantage must be in [0,1], got {target.advantage}")

    elif target.kind == "err_rates":
        if target.alpha is None or target.beta is None:
            raise ValueError("err_rates target requires both alpha and beta")
        if not 0 <= target.alpha <= 1:
            raise ValueError(f"alpha must be in [0,1], got {target.alpha}")
        if not 0 <= target.beta <= 1:
            raise ValueError(f"beta must be in [0,1], got {target.beta}")
        if target.alpha + target.beta >= 1:
            raise ValueError(
                f"Invalid trade-off: alpha + beta = {target.alpha + target.beta} >= 1"
            )

    elif target.kind == "epsilon_delta":
        if target.epsilon is None or target.delta is None:
            raise ValueError("epsilon_delta target requires both epsilon and delta")
        if target.epsilon < 0:
            raise ValueError(f"epsilon must be non-negative, got {target.epsilon}")
        if not 0 <= target.delta <= 1:
            raise ValueError(f"delta must be in [0,1], got {target.delta}")



"""DEPRECATED: Use riskcal.calibration.blackbox instead.

This module is deprecated and will be removed in v2.0.0.

For accountant-based calibration functions, use:
    from riskcal.calibration.blackbox import (
        find_noise_multiplier_for_epsilon_delta,
        find_noise_multiplier_for_advantage,
        find_noise_multiplier_for_err_rates,
    )
"""

import warnings

warnings.warn(
    "The 'riskcal.blackbox' module is deprecated and will be removed in v2.0.0. "
    "Use 'riskcal.calibration.blackbox' instead.",
    DeprecationWarning,
    stacklevel=2,
)

# Re-export from new locations for backward compatibility
from riskcal.calibration.blackbox import (
    find_noise_multiplier_for_epsilon_delta,
    find_noise_multiplier_for_advantage,
    find_noise_multiplier_for_err_rates,
    CalibrationResult,
)

__all__ = [
    "find_noise_multiplier_for_epsilon_delta",
    "find_noise_multiplier_for_advantage",
    "find_noise_multiplier_for_err_rates",
    "CalibrationResult",
]



"""Privacy risk analysis module.

This module provides functions for computing privacy metrics from various
privacy representations (PLDs, Gaussian mechanisms, epsilon-delta parameters).
"""

# Export PLRVs dataclass for advanced users
from riskcal.analysis._plrv import PLRVs, get_beta as get_beta_from_plrvs

# Export all functions from conversions module
from riskcal.analysis.conversions import (
    # Internal conversions
    plrvs_from_pld,
    pld_to_plrvs,
    # Beta functions
    get_beta_from_pld,
    get_beta_from_gdp,
    get_beta_from_rdp,
    get_beta_from_zcdp,
    get_beta_from_adp,
    get_beta_for_mu,
    get_beta_for_epsilon_delta,
    # Advantage functions
    get_advantage_from_pld,
    get_advantage_from_gdp,
    get_advantage_from_adp,
    get_advantage_from_zcdp,
    get_mu_from_zcdp_approx,
    get_beta_from_zcdp_approx,
    get_advantage_from_zcdp_approx,
    get_advantage_for_mu,
    get_advantage_for_epsilon_delta,
    # Bayes risk functions
    get_bayes_risk_from_pld,
    get_bayes_risk_from_gdp,
    get_bayes_risk_for_mu,
    # ADP
    get_epsilon_from_err_rates,
    get_epsilon_for_err_rates,
)

__all__ = [
    # Data structures
    "PLRVs",
    # Internal conversions
    "plrvs_from_pld",
    "pld_to_plrvs",
    # Beta (FNR) functions
    "get_beta_from_pld",
    "get_beta_from_gdp",
    "get_beta_from_adp",
    "get_beta_from_rdp",
    "get_beta_from_zcdp",
    "get_beta_for_mu",  # Deprecated alias
    "get_beta_for_epsilon_delta",  # Deprecated alias
    "get_beta_from_plrvs",
    # Advantage functions
    "get_advantage_from_pld",
    "get_advantage_from_gdp",
    "get_advantage_from_adp",
    "get_advantage_from_zcdp",
    "get_mu_from_zcdp_approx",  # Deprecated
    "get_beta_from_zcdp_approx",  # Deprecated
    "get_advantage_from_zcdp_approx",  # Deprecated
    "get_advantage_for_mu",  # Deprecated alias
    "get_advantage_for_epsilon_delta",  # Deprecated alias
    # Bayes risk functions
    "get_bayes_risk_from_pld",
    "get_bayes_risk_from_gdp",
    "get_bayes_risk_for_mu",  # Deprecated alias
    # Conversions
    "get_epsilon_from_err_rates",
    "get_epsilon_for_err_rates",  # Deprecated alias
]



"""DEPRECATED: Use riskcal.analysis instead.

This module is deprecated and will be removed in v2.0.0.

For risk/metric functions, use:
    from riskcal.analysis import get_beta_from_pld, get_advantage_from_pld, ...

For conversions, use:
    from riskcal.analysis import plrvs_from_pld, get_epsilon_for_err_rates
"""

import warnings

warnings.warn(
    "The 'riskcal.conversions' module is deprecated and will be removed in v2.0.0. "
    "Use 'riskcal.analysis' instead.",
    DeprecationWarning,
    stacklevel=2,
)

# Re-export from new locations for backward compatibility
from riskcal.analysis.conversions import (
    plrvs_from_pld,
    get_epsilon_for_err_rates,
    get_beta_from_pld,
    get_beta_for_mu,
    get_beta_for_epsilon_delta,
    get_advantage_from_pld,
    get_advantage_for_mu,
    get_advantage_for_epsilon_delta,
    get_bayes_risk_from_pld,
    get_bayes_risk_for_mu,
)

__all__ = [
    "plrvs_from_pld",
    "get_epsilon_for_err_rates",
    "get_beta_from_pld",
    "get_beta_for_mu",
    "get_beta_for_epsilon_delta",
    "get_advantage_from_pld",
    "get_advantage_for_mu",
    "get_advantage_for_epsilon_delta",
    "get_bayes_risk_from_pld",
    "get_bayes_risk_for_mu",
]



"""riskcal: Privacy risk analysis and calibration for differential privacy.

This library provides tools for computing f-DP trade-off curves for differential
privacy mechanisms, and calibrating their noise to operational privacy risk
(attack accuracy/advantage, or attack TPR and FPR) instead of (epsilon, delta).

Recommended usage:
    from riskcal import analysis  # For computing privacy metrics
    from riskcal import calibration  # For noise calibration
    from riskcal import accountants  # For privacy accountants
"""

# New recommended API
from riskcal import analysis
from riskcal import calibration
from riskcal import accountants

# Lazy deprecation for legacy modules and top-level names (PEP 562).
# Warnings are emitted only when deprecated names are actually accessed.

_DEPRECATED_MODULES = {
    "dpsgd": (
        "The 'riskcal.dpsgd' module is deprecated and will be removed in v2.0.0. "
        "Use 'riskcal.calibration.dpsgd' for calibration functions and "
        "'riskcal.accountants' for CTDAccountant instead."
    ),
    "conversions": (
        "The 'riskcal.conversions' module is deprecated and will be removed in v2.0.0. "
        "Use 'riskcal.analysis' instead."
    ),
    "blackbox": (
        "The 'riskcal.blackbox' module is deprecated and will be removed in v2.0.0. "
        "Use 'riskcal.calibration.blackbox' instead."
    ),
    "plrv": (
        "The 'riskcal.plrv' module is deprecated and will be removed in v2.0.0. "
        "Use 'riskcal.analysis' instead."
    ),
}

_DEPRECATED_TOPLEVEL = {
    # (new_module_path, migration_message)
    "get_advantage_from_pld": (
        "riskcal.analysis",
        "Use 'from riskcal.analysis import get_advantage_from_pld' instead.",
    ),
    "get_beta_from_pld": (
        "riskcal.analysis",
        "Use 'from riskcal.analysis import get_beta_from_pld' instead.",
    ),
    "get_advantage_for_dpsgd": (
        "riskcal.calibration.dpsgd",
        "Use 'from riskcal.calibration.dpsgd import get_advantage_for_dpsgd' instead.",
    ),
    "get_beta_for_dpsgd": (
        "riskcal.calibration.dpsgd",
        "Use 'from riskcal.calibration.dpsgd import get_beta_for_dpsgd' instead.",
    ),
    "find_noise_multiplier_for_advantage": (
        "riskcal.calibration.dpsgd",
        "Use 'from riskcal.calibration.dpsgd import find_noise_multiplier_for_advantage' instead.",
    ),
    "find_noise_multiplier_for_err_rates": (
        "riskcal.calibration.dpsgd",
        "Use 'from riskcal.calibration.dpsgd import find_noise_multiplier_for_err_rates' instead.",
    ),
    "CTDAccountant": (
        "riskcal.accountants",
        "Use 'from riskcal.accountants import CTDAccountant' instead.",
    ),
}


def __getattr__(name):
    import importlib
    import warnings

    if name in _DEPRECATED_MODULES:
        warnings.warn(_DEPRECATED_MODULES[name], DeprecationWarning, stacklevel=2)
        # Suppress the module-level warning to avoid double warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return importlib.import_module(f"riskcal.{name}")

    if name in _DEPRECATED_TOPLEVEL:
        module_path, msg = _DEPRECATED_TOPLEVEL[name]
        warnings.warn(
            f"Accessing 'riskcal.{name}' is deprecated and will be removed in v2.0.0. {msg}",
            DeprecationWarning,
            stacklevel=2,
        )
        mod = importlib.import_module(module_path)
        return getattr(mod, name)

    raise AttributeError(f"module 'riskcal' has no attribute {name!r}")


__all__ = [
    # New API (recommended)
    "analysis",
    "calibration",
    "accountants",
    # Legacy modules (deprecated, lazy)
    "dpsgd",
    "conversions",
    "blackbox",
    "plrv",
    # Legacy top-level exports (deprecated, lazy)
    "get_advantage_from_pld",
    "get_beta_from_pld",
    "get_advantage_for_dpsgd",
    "get_beta_for_dpsgd",
    "find_noise_multiplier_for_advantage",
    "find_noise_multiplier_for_err_rates",
    "CTDAccountant",
]



"""Privacy parameter calibration module.

This module provides tools for calibrating privacy parameters (such as noise
multipliers) to achieve target privacy guarantees specified in terms of
operational risk metrics (advantage, FPR/FNR) or traditional (epsilon, delta)-DP.
"""

# Export core calibration interface
from riskcal.calibration.core import (
    PrivacyEvaluator,
    PrivacyMetrics,
    CalibrationTarget,
    CalibrationConfig,
    CalibrationResult,
    calibrate_parameter,
)

# Export DP-SGD specific calibration
from riskcal.calibration.dpsgd import (
    create_dpsgd_evaluator,
    create_dpsgd_epsilon_evaluator,
    find_noise_multiplier_for_advantage as find_noise_multiplier_for_advantage_dpsgd,
    find_noise_multiplier_for_err_rates as find_noise_multiplier_for_err_rates_dpsgd,
    get_advantage_for_dpsgd,
    get_beta_for_dpsgd,
)

# Export blackbox/accountant-based calibration
from riskcal.calibration.blackbox import (
    create_accountant_evaluator,
    find_noise_multiplier_for_epsilon_delta,
    find_noise_multiplier_for_advantage as find_noise_multiplier_for_advantage_blackbox,
    find_noise_multiplier_for_err_rates as find_noise_multiplier_for_err_rates_blackbox,
    CalibrationResult as BlackboxCalibrationResult,  # Legacy format
)

__all__ = [
    # Core calibration interface
    "PrivacyEvaluator",
    "PrivacyMetrics",
    "CalibrationTarget",
    "CalibrationConfig",
    "CalibrationResult",
    "calibrate_parameter",
    # DP-SGD
    "create_dpsgd_evaluator",
    "create_dpsgd_epsilon_evaluator",
    "find_noise_multiplier_for_advantage_dpsgd",
    "find_noise_multiplier_for_err_rates_dpsgd",
    "get_advantage_for_dpsgd",
    "get_beta_for_dpsgd",
    # Blackbox/Accountant
    "create_accountant_evaluator",
    "find_noise_multiplier_for_epsilon_delta",
    "find_noise_multiplier_for_advantage_blackbox",
    "find_noise_multiplier_for_err_rates_blackbox",
    "BlackboxCalibrationResult",
]



"""Connect the Dots (CTD) accountant for DP-SGD."""

from copy import deepcopy
from dp_accounting.pld.privacy_loss_distribution import from_gaussian_mechanism


class CTDAccountant:
    """
    Opacus-compatible Connect the Dots accountant.

    This accountant tracks privacy loss for DP-SGD using Google's Connect the Dots
    (CTD) method via the dp_accounting library. It maintains a history of
    (noise_multiplier, sample_rate, num_steps) tuples and composes them to
    compute overall privacy loss.

    Example:
        >>> acct = CTDAccountant()
        >>> for _ in range(1000):
        ...     acct.step(noise_multiplier=1.0, sample_rate=0.01)
        >>> epsilon = acct.get_epsilon(delta=1e-5)
        >>> print(f"Privacy: ({epsilon:.2f}, 1e-5)-DP")
    """

    def __init__(self):
        self.history = []

    def step(self, *, noise_multiplier, sample_rate):
        """
        Record a single DP-SGD step.

        Args:
            noise_multiplier: Noise scale for this step.
            sample_rate: Poisson sampling probability for this step.
        """
        if len(self.history) >= 1:
            last_noise_multiplier, last_sample_rate, num_steps = self.history.pop()
            if (
                last_noise_multiplier == noise_multiplier
                and last_sample_rate == sample_rate
            ):
                self.history.append(
                    (last_noise_multiplier, last_sample_rate, num_steps + 1)
                )
            else:
                self.history.append(
                    (last_noise_multiplier, last_sample_rate, num_steps)
                )
                self.history.append((noise_multiplier, sample_rate, 1))

        else:
            self.history.append((noise_multiplier, sample_rate, 1))

    def get_pld(self, grid_step=1e-4, use_connect_dots=True):
        """
        Get the composed privacy loss distribution.

        Args:
            grid_step: Discretization interval for PLD computation.
            use_connect_dots: Whether to use Connect the Dots composition.

        Returns:
            Composed PrivacyLossDistribution from dp_accounting.
        """
        noise_multiplier, sample_rate, num_steps = self.history[0]
        pld = from_gaussian_mechanism(
            standard_deviation=noise_multiplier,
            sampling_prob=sample_rate,
            use_connect_dots=use_connect_dots,
            value_discretization_interval=grid_step,
        ).self_compose(num_steps)

        for noise_multiplier, sample_rate, num_steps in self.history[1:]:
            pld_new = from_gaussian_mechanism(
                standard_deviation=noise_multiplier,
                sampling_prob=sample_rate,
                use_connect_dots=use_connect_dots,
                value_discretization_interval=grid_step,
            ).self_compose(num_steps)
            pld = pld.compose(pld_new)

        return pld

    def get_epsilon(self, *, delta, **kwargs):
        """
        Get epsilon for given delta.

        Args:
            delta: Target delta value.
            **kwargs: Additional arguments passed to get_pld (e.g., grid_step).

        Returns:
            Epsilon value for (epsilon, delta)-DP.
        """
        pld = self.get_pld(**kwargs)
        return pld.get_epsilon_for_delta(delta)

    def get_beta(self, *, alpha, **kwargs):
        """
        Get FNR (beta) for given FPR (alpha).

        Args:
            alpha: False positive rate.
            **kwargs: Additional arguments passed to get_pld (e.g., grid_step).

        Returns:
            False negative rate corresponding to alpha.
        """
        from riskcal.analysis import get_beta_from_pld

        pld = self.get_pld(**kwargs)
        return get_beta_from_pld(pld, alpha)

    def get_advantage(self, **kwargs):
        """
        Get attack advantage.

        Args:
            **kwargs: Additional arguments passed to get_pld (e.g., grid_step).

        Returns:
            Maximum attack advantage.
        """
        from riskcal.analysis import get_advantage_from_pld

        pld = self.get_pld(**kwargs)
        return get_advantage_from_pld(pld)

    def __len__(self):
        """Return total number of steps recorded."""
        total = 0
        for _, _, steps in self.history:
            total += steps
        return total

    @classmethod
    def mechanism(cls):
        """Return mechanism name for state dict compatibility."""
        return "ctd"

    # The following methods are copied from https://opacus.ai/api/_modules/opacus/accountants/accountant.html#IAccountant
    # to avoid the direct dependence on the opacus package.

    def get_optimizer_hook_fn(self, sample_rate: float):
        """
        Returns a callback function which can be attached to DPOptimizer.

        Args:
            sample_rate: Expected sampling rate used for accounting.

        Returns:
            Hook function for DPOptimizer.
        """

        def hook_fn(optim):
            # This works for Poisson for both single-node and distributed
            # The reason is that the sample rate is the same in both cases (but in
            # distributed mode, each node samples among a subset of the data)
            self.step(
                noise_multiplier=optim.noise_multiplier,
                sample_rate=sample_rate * optim.accumulated_iterations,
            )

        return hook_fn

    def state_dict(self, destination=None):
        """
        Returns a dictionary containing the state of the accountant.

        Args:
            destination: A mappable object to populate the current state_dict into.
                If this arg is None, an OrderedDict is created and populated.
                Default: None.

        Returns:
            State dictionary.
        """
        if destination is None:
            destination = {}
        destination["history"] = deepcopy(self.history)
        destination["mechanism"] = self.mechanism()
        return destination

    def load_state_dict(self, state_dict):
        """
        Validates the supplied state_dict and populates the current
        Privacy Accountant's state dict.

        Args:
            state_dict: State dict to load.

        Raises:
            ValueError: If supplied state_dict is invalid and cannot be loaded.
        """
        if state_dict is None or len(state_dict) == 0:
            raise ValueError(
                "state dict is either None or empty and hence cannot be loaded"
                " into Privacy Accountant."
            )
        if "history" not in state_dict.keys():
            raise ValueError(
                "state_dict does not have the key `history`."
                " Cannot be loaded into Privacy Accountant."
            )
        if "mechanism" not in state_dict.keys():
            raise ValueError(
                "state_dict does not have the key `mechanism`."
                " Cannot be loaded into Privacy Accountant."
            )
        if self.mechanism() != state_dict["mechanism"]:
            raise ValueError(
                f"state_dict of {state_dict['mechanism']} cannot be loaded into "
                f" Privacy Accountant with mechanism {self.mechanism()}"
            )
        self.history = state_dict["history"]



"""
This module provides functions that compute a tradeoff curve T(P,Q)
given a PLRVs instance (Probability Loss Random Variables). We rely
on the notation from:

  - Kulynych et al. (https://arxiv.org/abs/2407.02191)
  - Dong et al. (https://arxiv.org/abs/1905.02383)

Below, we summarize the main concepts and notation needed to
understand how these functions work.

----------------------------------------------------------------------
DOMINATING PAIRS AND TRADEOFF FUNCTIONS
----------------------------------------------------------------------

Given a mechanism M, we say (P, Q) is a discrete-valued dominating
pair if, for all 0 <= alpha <= 1,

    T(P,Q)(alpha) <= T(M(D), M(D'))(alpha),

where T denotes the tradeoff function. (In this code, T(P,Q) is often
called f.)

We define random variables X and Y via:

    Y = log[P(o) / Q(o)]   with o ~ P,
    X = log[P(o') / Q(o')] with o' ~ Q.

In cases where (P, Q) have disjoint support, X can have a point mass
at -∞, and Y can have a point mass at +∞. Their domains are:

    Domain_X = {-∞} ∪ {x_0, x_1, ..., x_{k-1}}
             = {x_{-1}} ∪ {x_0, ..., x_{k-1}},

    Domain_Y = {y_0, y_1, ..., y_{l-1}} ∪ {+∞}
             = {y_0, ..., y_{l-1}} ∪ {y_l}.

----------------------------------------------------------------------
GOOGLE DP_ACCOUNTING LIBRARY AND DISCRETIZATION
----------------------------------------------------------------------

In practice, all PLRVs passed to this module come from the
`dp_accounting` library developed by Google. That library discretizes
the privacy loss random variables so that

    x_i = Δ * (x_0 + i), 0 <= i <= k - 1
    y_j = Δ * (y_0 + j), 0 <= j <= l - 1,

where the scalar Δ is the discretization parameter.
As the functions in this module are scale-invarient to Δ
(see, e.g. Algorithm 1 in Kulynych et al.), we let Δ=1 throughout
this module. As a result, the finite domain of X and Y takes on
equally spaced integer points, plus possible point masses at
-∞ (for X) and +∞ (for Y).

----------------------------------------------------------------------
PIECEWISE LINEARITY AND ALPHA_BAR
----------------------------------------------------------------------

We let f = T(P,Q). In some cases, f(alpha) != f^{-1}(alpha). This
discrepancy often arises under Poisson subsampling, where one order
corresponds to remove-adjacent datasets (D → D') and the other to
add-adjacent datasets (D' → D). When this occurs, we must apply a
symmetrization step to obtain the symmetric tradeoff curve that
matches the add/remove neighboring relation. This is precisely the
process described in Definition F.1 of Dong et al. (arXiv:1905.02383).

To carry out this symmetrization, we need to find the smallest value
of alpha for which -1 lies in the subdifferential of f. We denote
this value alpha_bar.

It can be shown that f is piecewise linear with breakpoints at
(Pr[X > x_i], Pr[Y <= x_i]) for -1 ≤ i ≤ k - 1. The linear segment
to the right of breakpoint i has slope -e^(x_i). At each breakpoint
i < k - 1, the subdifferential is the interval
[-e^(x_{i+1}), -e^(x_i)].

Let m be such that x_m = 0 (i.e., m = -x_0). Then -1 appears in
subdifferentials m and m - 1:

    [-e^(x_{m+1}), -e^(x_m)]
    [-e^(x_m), -e^(x_{m-1})]

Noting that alpha decreases as the index increases, the smallest
alpha at which -1 appears in the subdifferential corresponds to i = m.
This yields:

    alpha_bar = Pr[X > x_m] = Pr[X > 0],
    f(alpha_bar) = Pr[Y <= x_m] = Pr[Y <= 0].

Once alpha_bar is determined, implementing the symmetrization process
(Definition F.1 of Dong et al.) is straightforward.


----------------------------------------------------------------------
IMPLEMENTATIONS
----------------------------------------------------------------------

The file defines and uses:

  - `_tradeoff_function(plrvs, alphas)`:
      Computes T(P,Q)(alpha) (the FNR at each FPR alpha).

  - `_inverse_tradeoff_function(plrvs, alphas)`:
      Computes T(Q,P)(alpha), effectively the 'inverse' curve.

  - `get_beta(plrvs, alphas)`:
      Main entry point. Checks whether (P,Q) is symmetric; if not,
      computes a symmetrized tradeoff curve using alpha_bar, as above.

All of the above functions assume a PLRVs (defined below) instance as
input and rely on NumPy vectorization for efficient computation.
"""

import warnings
import numpy as np

from typing import Union
from dataclasses import dataclass

from dp_accounting.pld.privacy_loss_distribution import PrivacyLossDistribution

from riskcal.utils import _ensure_array


@dataclass
class PLRVs:
    """
    Privacy loss random variables.
    """

    y0: int
    pmf_Y: np.ndarray
    infinity_mass_Y: float

    x0: int
    pmf_X: np.ndarray
    minus_infinity_mass_X: float

    is_symmetric: bool


def _tradeoff_function(
    plrvs: PLRVs,
    alpha: np.ndarray,
) -> np.ndarray:
    """

    Computes the False Negative Rates (FNR) for a set of
    False Positive Rates (FPR), given the PLRVs X and Y.

    This function corresponds to Algorithm 1 in
    Kulynych et al. For each value of FPR, we find the
    Neyman-Pearson threshold in Domain_X, along with
    the Neyman-Pearson coin flip probability, then compute
    the corresponding FNR in Domain_Y.

    Args:

        plrvs:
            A PLRVs instance with attributes:
            - pmf_X, pmf_Y: Probability mass functions for X, Y
              over finite domains.
            - x0, y0: Offsets for the domain points.

        alpha: np.ndarray
            Ascending array of FPR values for which to compute FNR.

    Returns:
        beta: np.ndarray
            Array of FNR values corresponding to the input alpha.
    """

    # Reverse alpha for implementation convenience
    alpha = alpha[::-1]

    # Preallocate beta array for output
    beta = np.empty_like(alpha)

    # Length of domains of X and Y (excluding masses at +- ∞)
    k = len(plrvs.pmf_X)
    l = len(plrvs.pmf_Y)

    # ----------------------------------------------------------------------
    # Step 1: Construct complement CDF of X so that ccdf_X[i] = Pr[X > x_{i-1}],
    #         for 0 <= i <= k,
    #         where ccdf_X[0] = Pr[X > x_{-1}] = Pr[X > -∞] = 1 - Pr[X = -∞].
    #         The input pmf is over x_0, x_1, ..., x_{k-1}.
    # ----------------------------------------------------------------------

    ccdf_X = np.empty(k + 1)

    # Accumulate from the right: cumsum of reversed pmf gives us
    # the tail distribution. Then reverse again to get the usual order.
    ccdf_X[:-1] = np.cumsum(plrvs.pmf_X[::-1])[::-1]

    # Pr[X > x_{k-1}] = 0, so we add that as the last element
    ccdf_X[-1] = 0

    # ----------------------------------------------------------------------
    # Step 2: Construct cdf_Y so that cdf_Y[i] = Pr[Y <= y_i],
    #         for 0 <= i <= l - 1. Excludes the mass at +∞.
    #         The input pmf is over y_0, y_1, ..., y_{l-1}.
    # ----------------------------------------------------------------------

    cdf_Y = np.cumsum(plrvs.pmf_Y)

    # ----------------------------------------------------------------------
    # Step 3: Compute t array, where t[i] represents the index in X's domain
    #         of the NP threshold for FPR alpha[i].
    #         t ∈ [-1, k-1], with t = -1 meaning the threshold is -∞.
    # ----------------------------------------------------------------------

    # Search in ccdf_X[::-1] for the largest index where ccdf_X[index] <= alpha.
    # Note that reversing ccdf_X aligns with reversed alpha.
    t = len(plrvs.pmf_X) - np.searchsorted(ccdf_X[::-1], alpha, side="right")

    # ----------------------------------------------------------------------
    # Step 4: Compute j array for Y, which aligns threshold x_t to y_j.
    #         x_0 + t = y_0 + j
    # ----------------------------------------------------------------------

    j = plrvs.x0 + t - plrvs.y0

    # ----------------------------------------------------------------------
    # Step 5: Identify the four cases. See individual comments for details.
    # ----------------------------------------------------------------------

    # Case 1: t = -1 --> NP threshold = -infinity --> beta = 0
    index_case1 = np.searchsorted(
        t, -1, side="right"
    )  # all t behind index_case1 are <= -1

    # Case 2: j <= -1 --> NP threshold < y_0 --> beta = 0
    index_case2 = np.searchsorted(
        j, -1, side="right"
    )  # all j behind index_case2 are <= -1

    # handle Case 1 and 2 jointly.
    index_case12 = max(index_case1, index_case2)
    beta[:index_case12] = 0

    # Case 4: j > l - 1--> threshold > y_{l-1} --> beta = 1 - Pr[Y = inf]
    index_case4 = np.searchsorted(
        j, l - 1, side="right"
    )  # all j behind index_case4 are <= l - 1
    beta[index_case4:] = 1 - plrvs.infinity_mass_Y

    # Case 3: y_0 <= threshold <= y_{l-1}
    slice_case3 = np.index_exp[
        index_case12:index_case4
    ]  # indices of the thresholds in Case 3
    t_case3 = t[slice_case3]  # grab Case 3 threshold indices for Domain_X
    j_case3 = j[slice_case3]  # grab Case 3 threshold indices for Domain_Y

    # Note: use cdf_X[t_case3 + 1] because we want Pr[X <= x_{t_case3}]
    gammas = (alpha[slice_case3] - ccdf_X[t_case3 + 1]) / plrvs.pmf_X[t_case3]
    beta[slice_case3] = cdf_Y[j_case3] - gammas * plrvs.pmf_Y[j_case3]

    # reverse beta to align with input ascending alpha
    return beta[::-1]


def _inverse_tradeoff_function(
    plrvs: PLRVs,
    beta: np.ndarray,
) -> np.ndarray:
    """
    Computes the False Positive Rates (FPR) for a set of
    False Negative Rates (FNR), given the PLRVs X and Y.

    This function corresponds to the inverse of Algorithm 1 in
    Kulynych et al. For each value of FNR,  we find the
    Neyman-Pearson threshold in Domain_Y, along with
    the Neyman-Pearson coin flip probability, then compute
    the corresponding FNR in Domain_Y.

    Args:
        plrvs:
            A PLRVs instance with attributes:
            - pmf_X, pmf_Y: Probability mass functions for X, Y
              over finite domains.
            - x0, y0: Offsets for the domain points.
            - is_symmetric: Boolean indicating whether T(P,Q) = T(Q,P).

        beta: np.ndarray
            Ascending array of FNR values for which to compute FPR.

    Returns:
        alphas: np.ndarray
            Array of FPR values corresponding to the input beta.
    """

    # Preallocate alpha array for output
    alphas = np.empty_like(beta)

    # Length of domains of X and Y (excluding masses at +- inf)
    k = len(plrvs.pmf_X)
    l = len(plrvs.pmf_Y)

    # ----------------------------------------------------------------------
    # Step 1: Construct CDF of Y so that cdf_Y[i] = Pr[Y <= y_i],
    #         for 0 <= i <= l,
    #         where cdf_Y[l] = Pr[Y <= y_l] = Pr[Y <= ∞] = 1.
    #         The input pmf is over y_0, y_1, ..., y_{l-1}.
    # ----------------------------------------------------------------------
    cdf_Y = np.empty(l + 1)
    cdf_Y[:-1] = np.cumsum(plrvs.pmf_Y)
    cdf_Y[-1] = 1

    # ----------------------------------------------------------------------
    # Step 2: Construct compliment cdf of X so that ccdf_X[i] = Pr[X > x_i],
    #         for 0 <= i <= k - 1.
    # ----------------------------------------------------------------------
    ccdf_X = np.cumsum(plrvs.pmf_X[::-1])[::-1]
    ccdf_X = np.roll(ccdf_X, -1)
    ccdf_X[-1] = 0

    # ----------------------------------------------------------------------
    # Step 3: Compute t array, where t[i] represents the index in Y's domain
    #         of the NP threshold for FNR beta[i].
    #         t ∈ [0, l], with t = l meaning the threshold is ∞.
    # ----------------------------------------------------------------------

    # Search in cdf_Y for the first index where cdf_Y[index] >= beta.
    t = np.searchsorted(cdf_Y, beta, side="left")

    # ----------------------------------------------------------------------
    # Step 4: Compute j array for X, which aligns threshold y_t to x_j.
    #         y_0 + t = x_0 + j
    # ----------------------------------------------------------------------
    j = plrvs.y0 + t - plrvs.x0

    # ----------------------------------------------------------------------
    # Step 5: Identify the four cases. See individual comments for details.
    # ----------------------------------------------------------------------

    # Case 1: j < 0 --> NP threshold < x_0 --> alpha = 1 - Pr[X = -inf]
    index_case1 = np.searchsorted(
        j, -1, side="right"
    )  # all j behind index_case1 are <= -1
    alphas[:index_case1] = 1 - plrvs.minus_infinity_mass_X

    # Case 3: j > k - 1 --> threshold > x_k --> alpha = 0
    index_case3 = np.searchsorted(
        j, k - 1, side="right"
    )  # all j behind index_case3 are <= k - 1

    # Case 4: t = l --> threshold = infinity --> alpha = 0
    index_case4 = np.searchsorted(
        t, l - 1, side="right"
    )  # all t behind index_case4 are <= l - 1

    # handle Case 3 and 4 jointly.
    index_case34 = min(index_case3, index_case4)
    alphas[index_case34:] = 0

    # Case 2: x_0 <= threshold <= x_{l-1}
    slice_case2 = np.index_exp[
        index_case1:index_case34
    ]  # indices of the thresholds in Case 3
    t_case2 = t[slice_case2]  # grab Case 2 threshold indices for Domain_Y
    j_case2 = j[slice_case2]  # grab Case 2 threshold indices for Domain_X
    gammas = (cdf_Y[t_case2] - beta[slice_case2]) / plrvs.pmf_Y[t_case2]
    alphas[slice_case2] = ccdf_X[j_case2] + gammas * plrvs.pmf_X[j_case2]

    # Done! Return alphas
    return alphas


def get_beta(
    plrvs: PLRVs,
    alpha: Union[float, np.ndarray] = None,
    alphas: Union[float, np.ndarray] = None,  # Deprecated.
) -> Union[float, np.ndarray]:
    """
    Computes the FNR values corresponding to the input FPR values (alphas).

    .. deprecated:: 1.2.0
        Parameter 'alphas' is deprecated and will be removed in version 2.0.0.
        Use 'alpha' instead.

    This function applies the following logic:

    1. Ensures the input alphas are in a NumPy array (float or 1D array).
       If they are not sorted, sort them in ascending order.

    2. Check whether the PLRVs object is symmetric. If so, call
       the direct tradeoff function T(P,Q) (i.e., _tradeoff_function).

    3. If not symmetric, apply the symmetrization logic of
       Definition F.1 from Dong et al. (see file docstring for details)

       a. Compute alpha_bar = Pr[X > 0] and f_alpha_bar = Pr[Y <= 0].

       b. Compare alpha_bar and f_alpha_bar:

          - If alpha_bar <= f_alpha_bar, the symmetrized tradeoff curve
            includes a region in which the curve is linear and equal to
            alpha_bar + f_alpha_bar - alpha. Outside that region,
            the curve follows _tradeoff_function on one side and
            _inverse_tradeoff_function on the other.

          - Otherwise, the symmetrized curve at each alpha is given by
            the maximum of T(P,Q)(alpha) and T(Q,P)(alpha).

    4. Returns the FNR values (beta) in the same shape (scalar or array)
       as the input alpha.

    Args:
        plrvs: A PLRVs instance with attributes:
            - pmf_X, pmf_Y: Probability mass functions for X, Y over finite domains.
            - x0, y0: Offsets for the domain points.
            - is_symmetric: Boolean indicating whether T(P,Q) = T(Q,P).
        alpha: A float or array of floats representing the False Positive Rates
            for which we need to compute the False Negative Rates.

    Returns:
        A float (if alpha was a single float) or a NumPy array (if alpha
        was an array) of False Negative Rates corresponding to the input alpha.
    """
    if alpha is None and alphas is None:
        raise ValueError("Must specify alpha.")

    elif alpha is not None and alphas is not None:
        raise ValueError("Must pass either alpha or alphas.")

    elif alphas is not None:
        warnings.warn(
            "Parameter 'alphas' is deprecated and will be removed in a future version. "
            "Use 'alpha' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        alpha = alphas

    # Convert alpha to array; check if input was scalar
    alpha, is_scalar = _ensure_array(alpha)

    # Ensure ascending order
    is_sorted = np.all(np.diff(alpha) >= 0)
    if not is_sorted:
        sort_idx = np.argsort(alpha)
        alpha = alpha[sort_idx]
        undo_sort_indices = np.argsort(sort_idx)

    # Symmetric means T(P,Q) = T(Q,P). No need to symmetrize
    if plrvs.is_symmetric:
        output = _tradeoff_function(plrvs, alpha)

    # See module docstring for explanation
    else:

        # ----------------------------------------------------------------------
        # Step a: Compute alpha_bar, f_alpha_bar
        #         alpha_bar = Pr[X > 0] = sum of pmf_X for all x > 0
        #         f_alpha_bar = Pr[Y <= 0] = sum of pmf_Y for all y <= 0
        #
        #         since pmf_X[i] = Pr[X = x_0 + i], it follows that
        #         pmf_X[-x0 + 1] = Pr[X = 1], hence we sum from there onwards.
        #         Similar logic holds for pmf_Y.
        # ----------------------------------------------------------------------

        alpha_bar = np.sum(plrvs.pmf_X[-plrvs.x0 + 1 :])
        f_alpha_bar = np.sum(plrvs.pmf_Y[: -plrvs.y0 + 1])

        # ----------------------------------------------------------------------
        # Step b: Apply Definition F.1
        # ----------------------------------------------------------------------
        if alpha_bar <= f_alpha_bar:

            # Region partitioning input alpha based on alpha_bar, f_alpha_bar
            alpha_bar_index = np.searchsorted(
                alpha, alpha_bar, side="right"
            )  # all alpha behind index are <= alpha_bar
            f_alpha_bar_index = np.searchsorted(
                alpha, f_alpha_bar, side="right"
            )  # all alpha behind index are <= f_alpha_bar

            output = np.empty_like(alpha)

            # 1) Evaluate the symmeterized tradeoff curve for alpha < alpha_bar
            output[:alpha_bar_index] = _tradeoff_function(
                plrvs, alpha[:alpha_bar_index]
            )

            # 2) Evaluate the symmeterized tradeoff curve in linear region alpha_bar <= alpha <= f_alpha_bar
            output[alpha_bar_index:f_alpha_bar_index] = (
                alpha_bar + f_alpha_bar - alpha[alpha_bar_index:f_alpha_bar_index]
            )

            # 3) Evaluate the symmeterized tradeoff curve for f_alpha_bar < alpha
            output[f_alpha_bar_index:] = _inverse_tradeoff_function(
                plrvs, alpha[f_alpha_bar_index:]
            )

        else:
            # When alpha_bar > f_alpha_bar, symmetrization = maximum of
            # T(P,Q)(alpha) and T(Q,P)(alpha).
            tradeoff_arr = _tradeoff_function(plrvs, alpha)
            inverse_tradeoff_arr = _inverse_tradeoff_function(plrvs, alpha)
            output = np.maximum(tradeoff_arr, inverse_tradeoff_arr)

    # Output item if input alpha was a scalar
    if is_scalar:
        return output.item()

    # If input array alpha were not sorted, undo the sort
    if not is_sorted:
        output = output[undo_sort_indices]

    # return numpy array
    return output



import numpy as np
import pytest

import riskcal
from opacus import accountants
from scipy.stats import norm
from scipy.optimize import root_scalar


@pytest.fixture(params=[accountants.rdp.RDPAccountant])
def accountant(request):
    return request.param


sample_rate = 0.01
num_dpsgd_steps = 100


@pytest.mark.parametrize(
    "advantage, sample_rate, num_steps",
    [
        # Gaussian mechanism.
        (0.01, 1, 1),
        (0.10, 1, 1),
        (0.25, 1, 1),
        # DP-SGD.
        (0.01, sample_rate, num_dpsgd_steps),
        (0.10, sample_rate, num_dpsgd_steps),
        (0.25, sample_rate, num_dpsgd_steps),
    ],
)
def test_adv_calibration_correctness(accountant, advantage, sample_rate, num_steps):

    # chosen ad hoc for speed
    rdp_alphas = np.linspace(1.5, 101, 50)

    epsilon_error = 1e-4
    calibrated_mu = riskcal.blackbox.find_noise_multiplier_for_advantage(
        accountant,
        advantage=advantage,
        sample_rate=sample_rate,
        num_steps=num_steps,
        eps_error=epsilon_error,
        alphas=rdp_alphas,
    )

    acct_obj = accountant()
    for _ in range(num_steps):
        acct_obj.step(noise_multiplier=calibrated_mu, sample_rate=sample_rate)

    # Verify that mu is calibrated for (0, adv)-DP:
    numerical_epsilon = acct_obj.get_epsilon(delta=advantage, alphas=rdp_alphas)
    assert pytest.approx(numerical_epsilon, abs=epsilon_error) == 0


@pytest.mark.parametrize(
    "beta, sample_rate, num_steps, method",
    [
        # Gaussian mechanism:
        (0.25, 1, 1, "bounded"),
        (0.50, 1, 1, "bounded"),
        (0.75, 1, 1, "bounded"),
        # DP-SGD
        (0.25, sample_rate, num_dpsgd_steps, "bounded"),
        (0.50, sample_rate, num_dpsgd_steps, "bounded"),
        (0.75, sample_rate, num_dpsgd_steps, "bounded"),
    ],
)
def test_err_rates_calibration_correctness(
    accountant, beta, sample_rate, num_steps, method
):
    alpha = 0.01
    classical_delta = 1e-5
    delta_error = 0.001
    eps_error = 0.01
    rdp_alphas = np.linspace(1 + 1e-2, 17, 31)
    calibration_result = riskcal.blackbox.find_noise_multiplier_for_err_rates(
        accountant,
        alpha=alpha,
        beta=beta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        delta_error=delta_error,
        alphas=rdp_alphas,
    )
    calibrated_mu = calibration_result.noise_multiplier
    calibrated_delta = calibration_result.calibration_delta

    acct_obj = accountant()
    for _ in range(num_steps):
        acct_obj.step(noise_multiplier=calibrated_mu, sample_rate=sample_rate)

    epsilon = acct_obj.get_epsilon(delta=calibrated_delta, alphas=rdp_alphas)
    expected_epsilon = calibration_result.calibration_epsilon

    print(f"CHECK 1: {alpha=}, {beta=} // {epsilon=}, {expected_epsilon=}")
    assert epsilon == pytest.approx(expected_epsilon, abs=eps_error)

    obtained_beta = riskcal.conversions.get_beta_for_epsilon_delta(
        epsilon, calibrated_delta, alpha
    )
    print(f"CHECK 2: {alpha=}, {beta=} // {epsilon=}, {obtained_beta=}")
    assert beta == pytest.approx(obtained_beta, abs=delta_error)


@pytest.mark.parametrize(
    "epsilon, sample_rate, num_steps",
    [
        (1.0, 1.0, 1),
        (4.0, 1.0, 1),
        (1.0, sample_rate, num_dpsgd_steps),
        (4.0, sample_rate, num_dpsgd_steps),
    ],
)
def test_err_rates_calibration_improvement(accountant, epsilon, sample_rate, num_steps):
    alpha = 0.01
    delta = 1e-5
    delta_error = 0.001
    method = "bounded"
    rdp_alphas = np.linspace(1 + 1e-4, 520, 45)

    standard_mu = riskcal.blackbox.find_noise_multiplier_for_epsilon_delta(
        accountant,
        epsilon=epsilon,
        delta=delta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        alphas=rdp_alphas,
    )

    # What is the FNR at alpha = 0.1 for the target epsilon?
    beta = riskcal.conversions.get_beta_for_epsilon_delta(epsilon, delta, alpha)
    # Let's calibrate directly for these alpha / beta:
    calibration_result = riskcal.blackbox.find_noise_multiplier_for_err_rates(
        accountant,
        alpha=alpha,
        beta=beta,
        sample_rate=sample_rate,
        num_steps=num_steps,
        delta_error=delta_error,
        method=method,
        alphas=rdp_alphas,
    )
    calibrated_mu = calibration_result.noise_multiplier
    calibrated_delta = calibration_result.calibration_delta

    # We should get less noise with direct calibration:
    assert standard_mu > calibrated_mu

    # Check that alpha beta guarantees are correct.
    acct_obj = accountant()
    for step in range(num_steps):
        acct_obj.step(noise_multiplier=calibrated_mu, sample_rate=sample_rate)
    obtained_epsilon = acct_obj.get_epsilon(delta=calibrated_delta, alphas=rdp_alphas)

    obtained_beta = riskcal.conversions.get_beta_for_epsilon_delta(
        obtained_epsilon, calibrated_delta, alpha
    )
    assert beta == pytest.approx(obtained_beta, abs=delta_error)


@pytest.mark.parametrize(
    "beta, method",
    [
        (0.25, "bounded"),
        (0.50, "bounded"),
        (0.75, "bounded"),
    ],
)
def test_generic_err_rates_calibration_worse_than_exact(beta, method):
    alpha = 0.1
    delta_error = 1e-2

    class GaussianMechanismAccountant:
        def __init__(self):
            pass

        def step(self, noise_multiplier: float, **kwargs):
            self.noise_multiplier = noise_multiplier

        def get_epsilon(self, delta):
            def get_delta(epsilon):
                return norm.cdf(
                    0.5 / self.noise_multiplier - epsilon * self.noise_multiplier
                ) - np.exp(epsilon) * norm.cdf(
                    -0.5 / self.noise_multiplier - epsilon * self.noise_multiplier
                )

            return root_scalar(lambda epsilon: get_delta(epsilon) - delta, x0=0.0).root

    calibration_result = riskcal.blackbox.find_noise_multiplier_for_err_rates(
        GaussianMechanismAccountant,
        alpha=alpha,
        beta=beta,
        sample_rate=1,
        num_steps=1,
        delta_error=delta_error,
        method=method,
    )

    exact_noise_multiplier = 1 / (norm.ppf(1 - alpha) - norm.ppf(beta))
    # assert exact_noise_multiplier <= calibration_result.noise_multiplier
    assert exact_noise_multiplier == pytest.approx(
        calibration_result.noise_multiplier, abs=delta_error
    )



import pytest
import numpy as np
import math

from scipy.optimize import minimize_scalar
from dp_accounting.pld.privacy_loss_distribution import from_gaussian_mechanism

import riskcal


def _get_bayes_risk_from_kaissis_definition(pld, prior):
    """Optimization-based Bayes risk. Kaissis et al. (2024), Eq. 2. https://arxiv.org/abs/2406.08918"""
    from riskcal.utils import _ensure_array

    prior, is_scalar = _ensure_array(prior)
    bayes_risk = []
    for prior_val in prior:
        result = minimize_scalar(
            lambda x: prior_val * x
            + (1 - prior_val) * riskcal.analysis.get_beta_from_pld(pld, alpha=x),
            bounds=(0, 1),
            method="bounded",
        )
        bayes_risk.append(result.fun)
    if is_scalar:
        return bayes_risk[0]
    return np.array(bayes_risk)


# Global tolerances for test assertions
REL_TOLERANCE = 1e-4  # Relative tolerance for numerical comparisons
ABS_TOLERANCE = 1e-12  # Absolute tolerance for small numbers


def get_gaussian_plrv_data():
    mu = 2.0
    pld = from_gaussian_mechanism(
        standard_deviation=1.0 / mu, value_discretization_interval=1e-3
    )
    return mu, pld, riskcal.analysis.conversions.plrvs_from_pld(pld)


def get_subsampled_gaussian_plrv_data():
    mu = 2.5
    pld = from_gaussian_mechanism(
        standard_deviation=1.0 / mu, sampling_prob=0.1
    ).self_compose(5)
    return mu, pld, riskcal.analysis.conversions.plrvs_from_pld(pld)


@pytest.mark.parametrize(
    "alpha",
    [
        0.1,
        np.array([0.1, 0.2, 0.3]),
    ],
)
def test_get_beta_matches_analytic_curve(alpha):
    mu, pld, plrvs = get_gaussian_plrv_data()
    analytic_beta = riskcal.analysis.conversions.get_beta_for_mu(mu, alpha)
    numerical_beta1 = riskcal.plrv.get_beta(plrvs, alpha=alpha)
    numerical_beta2 = riskcal.analysis.conversions.get_beta_from_pld(pld, alpha=alpha)

    # check that numerical beta is close to analytic beta
    assert analytic_beta == pytest.approx(
        numerical_beta1, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )
    assert analytic_beta == pytest.approx(
        numerical_beta2, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


def test_deprecated_get_beta_interface():
    mu, pld, plrvs = get_gaussian_plrv_data()
    alpha = 0.5
    analytic_beta = riskcal.analysis.conversions.get_beta_for_mu(mu, alpha)

    with pytest.warns(DeprecationWarning):
        numerical_beta1 = riskcal.plrv.get_beta(plrvs, alphas=alpha)
        numerical_beta2 = riskcal.analysis.conversions.get_beta_from_pld(
            pld, alphas=alpha
        )

    # check that numerical beta is close to analytic beta
    assert analytic_beta == pytest.approx(
        numerical_beta1, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )
    assert analytic_beta == pytest.approx(
        numerical_beta2, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


def test_get_beta_symmetrized_in_the_middle():
    mu, pld, plrvs = get_subsampled_gaussian_plrv_data()
    analytic_advantage = pld.get_delta_for_epsilon(0)
    analytic_fixed_point_beta = 0.5 * (1 - analytic_advantage)

    alpha = np.linspace(0, 1, 100_000)
    beta = riskcal.analysis.conversions.get_beta_from_pld(
        pld,
        alpha=alpha,
    )
    numerical_fixed_point_idx = np.argmin(np.abs(beta - analytic_fixed_point_beta))
    numerical_fixed_point = beta[numerical_fixed_point_idx]

    assert analytic_fixed_point_beta == pytest.approx(
        numerical_fixed_point, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


def test_get_advantage_matches_analytic_matching_expression():
    mu, pld, plrvs = get_gaussian_plrv_data()
    analytic_advantage = riskcal.analysis.conversions.get_advantage_for_mu(mu)
    numerical_advantage = riskcal.analysis.conversions.get_advantage_from_pld(pld)
    assert analytic_advantage == pytest.approx(
        numerical_advantage, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


@pytest.mark.parametrize(
    "prior",
    [
        0.5,
        np.array([0.1, 0.2, 0.3]),
    ],
)
def test_bayes_risk_from_gdp_matches_pld(prior):
    mu, pld, plrvs = get_gaussian_plrv_data()

    bayes_risk_from_pld = riskcal.analysis.conversions.get_bayes_risk_from_pld(
        pld, prior
    )
    bayes_risk_from_gdp = riskcal.analysis.conversions.get_bayes_risk_for_mu(mu, prior)
    assert bayes_risk_from_pld == pytest.approx(
        bayes_risk_from_gdp, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )

    # When the prior is 0.5, check that we correctly recover TV privacy.
    if not hasattr(prior, "__len__") and prior == 0.5:
        advantage = riskcal.analysis.conversions.get_advantage_from_gdp(mu)
        risk_from_advantage = 0.5 - 0.5 * advantage
        assert bayes_risk_from_pld == pytest.approx(
            risk_from_advantage, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
        )
        assert bayes_risk_from_gdp == pytest.approx(
            risk_from_advantage, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
        )


@pytest.mark.parametrize(
    "prior",
    [
        0.1,
        0.2,
        0.25,
        0.75,
        0.8,
        0.9,
    ],
)
def test_bayes_risk_is_symmetric(prior):
    mu, pld, plrvs = get_gaussian_plrv_data()

    numerical_risk1 = riskcal.analysis.conversions.get_bayes_risk_from_pld(pld, prior)
    numerical_risk2 = riskcal.analysis.conversions.get_bayes_risk_from_pld(
        pld, 1 - prior
    )
    assert numerical_risk1 == pytest.approx(
        numerical_risk2, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )

    analytic_risk1 = riskcal.analysis.conversions.get_bayes_risk_for_mu(mu, prior)
    analytic_risk2 = riskcal.analysis.conversions.get_bayes_risk_for_mu(mu, 1 - prior)
    assert analytic_risk1 == pytest.approx(
        analytic_risk2, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


@pytest.mark.parametrize(
    "prior",
    [
        0.5,
        np.array([0.1, 0.2, 0.3]),
        np.array([0.75, 0.8, 0.9]),
    ],
)
def test_bayes_risk_from_pld_matches_kaissis_definition(prior):
    mu, pld, plrvs = get_gaussian_plrv_data()

    closed_form_risk_from_pld = riskcal.analysis.conversions.get_bayes_risk_from_pld(
        pld, prior
    )
    closed_form_risk_from_gdp = riskcal.analysis.conversions.get_bayes_risk_from_gdp(
        mu, prior
    )
    kaissis_risk = _get_bayes_risk_from_kaissis_definition(pld, prior)
    assert closed_form_risk_from_pld == pytest.approx(
        kaissis_risk, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )
    assert closed_form_risk_from_gdp == pytest.approx(
        kaissis_risk, rel=REL_TOLERANCE, abs=ABS_TOLERANCE
    )


def test_advantage_from_zcdp_decreases_with_smaller_rho():
    """Test that advantage decreases as privacy increases (rho decreases)."""
    rhos = [2.0, 1.0, 0.5, 0.1]
    advantages = [riskcal.analysis.get_advantage_from_zcdp(rho=rho) for rho in rhos]

    # Advantages should be in decreasing order
    for i in range(len(advantages) - 1):
        assert advantages[i] > advantages[i + 1], (
            f"Advantage should decrease as rho decreases: "
            f"rho={rhos[i]} gave {advantages[i]}, "
            f"rho={rhos[i+1]} gave {advantages[i+1]}"
        )


# =============================================================================
# RDP (Renyi Differential Privacy)
# =============================================================================


@pytest.mark.parametrize(
    "epsilon,order",
    [
        (1.0, 2.0),   # order > 1
        (0.5, 1.0),   # order == 1
        (1.0, 0.7),   # order < 1
    ],
)
def test_get_beta_from_rdp_is_decreasing(epsilon, order):
    """Beta should be monotonically decreasing as alpha (FPR) increases."""
    alphas = np.linspace(0.01, 0.99, 50)
    betas = np.array([
        riskcal.analysis.get_beta_from_rdp(epsilon=epsilon, alpha=a, order=order)
        for a in alphas
    ])
    # Allow tolerance matching the bisection accuracy (tol=1e-7 default)
    assert np.all(np.diff(betas) <= 1e-6), (
        f"beta not monotonically decreasing for epsilon={epsilon}, order={order}"
    )


@pytest.mark.parametrize(
    "epsilon,order",
    [
        (1.0, 2.0),
        (0.5, 1.0),
        (1.0, 0.7),
    ],
)
def test_get_beta_from_rdp_scalar_and_array_agree(epsilon, order):
    """Scalar and array inputs should produce identical results."""
    alphas = np.array([0.01, 0.05, 0.1, 0.3, 0.5])
    beta_array = riskcal.analysis.get_beta_from_rdp(
        epsilon=epsilon, alpha=alphas, order=order
    )
    beta_scalars = np.array([
        riskcal.analysis.get_beta_from_rdp(epsilon=epsilon, alpha=float(a), order=order)
        for a in alphas
    ])
    np.testing.assert_allclose(beta_array, beta_scalars, rtol=1e-6)


@pytest.mark.parametrize("epsilon,order", [(1.0, 2.0), (0.5, 3.0), (2.0, 1.5)])
def test_get_beta_from_rdp_edge_cases(epsilon, order):
    """At alpha=0 beta should be 1; at alpha=1 beta should be 0."""
    assert riskcal.analysis.get_beta_from_rdp(
        epsilon=epsilon, alpha=0.0, order=order
    ) == pytest.approx(1.0, abs=1e-6)
    assert riskcal.analysis.get_beta_from_rdp(
        epsilon=epsilon, alpha=1.0, order=order
    ) == pytest.approx(0.0, abs=1e-6)


@pytest.mark.parametrize("rho", list(np.linspace(0.1, 3.0, 6)))
def test_get_beta_from_rdp_single_order_leq_zcdp(rho):
    """
    get_beta_from_zcdp maximizes over orders, so it must be >= any single-order bound.
    """
    order = 3.0
    alpha = 0.1
    beta_rdp = riskcal.analysis.get_beta_from_rdp(
        epsilon=rho * order, alpha=alpha, order=order
    )
    beta_zcdp = riskcal.analysis.get_beta_from_zcdp(rho=rho, alpha=alpha)
    assert beta_zcdp >= beta_rdp - 1e-6, (
        f"zCDP beta {beta_zcdp} should be >= single-order RDP beta {beta_rdp}"
    )


def test_get_beta_from_rdp_backward_compat_params():
    """linear_search_step and max_bisection_steps are accepted but ignored."""
    beta1 = riskcal.analysis.get_beta_from_rdp(epsilon=1.0, alpha=0.1, order=2.0)
    beta2 = riskcal.analysis.get_beta_from_rdp(
        epsilon=1.0, alpha=0.1, order=2.0,
        linear_search_step=0.01,
        max_bisection_steps=10,
    )
    assert beta1 == pytest.approx(beta2, rel=1e-9)
